from typing import Callable, List, Tuple, Union
from uuid import UUID, uuid4

from pydantic import BaseModel, Field, StrictFloat, StrictInt

from unfolded.map_sdk.api.enums import ActionType

Number = Union[StrictFloat, StrictInt]
Range = Tuple[Number, Number]
TimeRange = Tuple[Number, Number]


class ApiBaseModel(BaseModel):
    class Config:
        validate_assignment = True
        extra = "forbid"


class CamelCaseBaseModel(ApiBaseModel):
    class Config:
        allow_population_by_field_name = True

        @staticmethod
        def to_camel_case(snake_str: str) -> str:
            """Convert snake_case string to camelCase
            https://stackoverflow.com/a/19053800
            """
            components = snake_str.split("_")
            # We capitalize the first letter of each component except the first one
            # with the 'title' method and join them together.
            return components[0] + "".join(x.title() for x in components[1:])

        alias_generator = to_camel_case


class KebabCaseBaseModel(ApiBaseModel):
    class Config:
        allow_population_by_field_name = True
        alias_generator: Callable[[str], str] = lambda snake_str: snake_str.replace(
            "_", "-"
        )


class Action(CamelCaseBaseModel):
    """Base Action payload class"""

    class Meta:
        args: List[str] = []
        """Order in which arguments should be serialized"""

        options: List[str] = []
        """Fields to be collected into an options dict/object."""

    type: ActionType
    message_id: UUID = Field(default_factory=uuid4)



import json
from typing import List, Optional, Type
from uuid import uuid4

from jinja2 import Environment, PackageLoader

from unfolded.map_sdk._version import __version__
from unfolded.map_sdk.api.base import Action
from unfolded.map_sdk.transport.base import BaseNonInteractiveTransport
from unfolded.map_sdk.types import ResponseClass
from unfolded.map_sdk.utils.serialization import serialize_action


class HTMLTransport(BaseNonInteractiveTransport):
    """Transport used in a static HTML map"""

    action_list: List[Action]

    def __init__(self) -> None:
        super().__init__()
        self.action_list = []

    def send_action(
        self,
        *,
        action: Action,
        response_class: Optional[Type[ResponseClass]] = None,
    ) -> None:
        # pylint:disable=unused-argument
        self.action_list.append(action)

    @property
    def serialized_actions(self):
        return [
            serialize_action(action, renderer="html")[0] for action in self.action_list
        ]

    def render_template(self, **kwargs) -> str:
        self.rendered = True
        j2_env = Environment(
            loader=PackageLoader("unfolded.map_sdk"),
            autoescape=True,
            trim_blocks=True,
            keep_trailing_newline=True,
        )
        template = j2_env.get_template("html_map_sdk.j2")

        constructor_args = {key: json.dumps(val) for key, val in kwargs.items()}

        html_str = template.render(
            version=__version__,
            # We need to escape newlines again since the templating removes a backslash.
            # Since we control self.serialized actions, the only newlines will come from
            # pd.to_csv() or other user-provided input, which will be inside a string
            actions=json.dumps(self.serialized_actions)
            .replace(r"\n", r"\\n")
            .replace(r"\t", r"\\t")
            .replace(r"\"", r'\\"'),
            div_id=uuid4(),
            **constructor_args,
        )
        return html_str



from tests._utils import check_expected_comm_message
from tests.conftest import MockComm
from unfolded.map_sdk.map.widget import SyncWidgetMap


class TestGetLayers:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_layers()

        expected = {
            "type": "v1/map-sdk-get-layers",
            "args": [],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestGetLayerById:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_layer_by_id("layer-id")

        expected = {
            "type": "v1/map-sdk-get-layer-by-id",
            "args": ["layer-id"],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestAddLayer:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-add-layer",
            "args": [
                {"id": "layer-id", "dataId": "data-1", "fields": {"field-1": "value-1"}}
            ],
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        layer = {
            "id": "layer-id",
            "data_id": "data-1",
            "fields": {"field-1": "value-1"},
        }

        mock_widget_map.add_layer(layer)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        layer = {
            "id": "layer-id",
            "data_id": "data-1",
            "fields": {"field-1": "value-1"},
        }

        mock_widget_map.add_layer(**layer)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_add_layer_with_config(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        layer_id = "earthquake_points"
        dataset_id = "dataset-id"
        layer = {
            "id": layer_id,
            "type": "point",
            "data_id": dataset_id,
            "label": "Earthquakes",
            "is_visible": True,
            "fields": {"lat": "Latitude", "lng": "Longitude"},
            "config": {
                "visual_channels": {
                    "colorField": {"name": "Depth", "type": "real"},
                }
            },
        }
        mock_widget_map.add_layer(layer)

        expected = {
            "type": "v1/map-sdk-add-layer",
            "args": [
                {
                    "id": layer_id,
                    "type": "point",
                    "dataId": dataset_id,
                    "label": "Earthquakes",
                    "isVisible": True,
                    "fields": {"lat": "Latitude", "lng": "Longitude"},
                    "config": {
                        "visualChannels": {
                            "colorField": {"name": "Depth", "type": "real"},
                        }
                    },
                }
            ],
        }
        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestUpdateLayer:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-update-layer",
            "args": [
                "layer-id",
                {"fields": {"field-1": "value-1"}},
            ],
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        values = {"fields": {"field-1": "value-1"}}
        mock_widget_map.update_layer("layer-id", values)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        values = {"fields": {"field-1": "value-1"}}
        mock_widget_map.update_layer("layer-id", **values)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)


class TestRemoveLayer:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.remove_layer("layer-id")

        expected = {
            "type": "v1/map-sdk-remove-layer",
            "args": ["layer-id"],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestGetLayerGroups:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_layer_groups()

        expected = {
            "type": "v1/map-sdk-get-layer-groups",
            "args": [],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestGetLayerGroupById:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_layer_group_by_id("layer-group-id")

        expected = {
            "type": "v1/map-sdk-get-layer-group-by-id",
            "args": ["layer-group-id"],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestAddLayerGroupAction:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-add-layer-group",
            "args": [
                {"id": "layer-group-id", "label": "layer-group-1", "isVisible": False}
            ],
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        layer_group = {
            "id": "layer-group-id",
            "label": "layer-group-1",
            "is_visible": False,
        }
        mock_widget_map.add_layer_group(layer_group)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        layer_group = {
            "id": "layer-group-id",
            "label": "layer-group-1",
            "is_visible": False,
        }
        mock_widget_map.add_layer_group(**layer_group)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)


class TestUpdateLayerGroupAction:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-update-layer-group",
            "args": [
                "layer-group-id",
                {"label": "layer-group-2"},
            ],
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        values = {"label": "layer-group-2"}
        mock_widget_map.update_layer_group("layer-group-id", values)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        values = {"label": "layer-group-2"}
        mock_widget_map.update_layer_group("layer-group-id", **values)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)


class TestRemoveLayerGroup:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.remove_layer_group("layer-group-id")

        expected = {
            "type": "v1/map-sdk-remove-layer-group",
            "args": ["layer-group-id"],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestGetLayerTimeline:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_layer_timeline()

        expected = {
            "type": "v1/map-sdk-get-layer-timeline",
            "args": [],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestUpdateLayerTimeline:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-update-layer-timeline",
            "args": [{"currentTime": 0}],
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.update_layer_timeline({"current_time": 0})

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        mock_widget_map.update_layer_timeline(**{"current_time": 0})

        assert check_expected_comm_message(self.expected, mock_comm.log_send)



""" Block notebook cells from running while interacting with widgets
"""

from ._poll import run_ui_poll_loop, ui_events, with_ui_events  # type: ignore
from ._version import __version__

__all__ = (
    "ui_events",
    "with_ui_events",
    "run_ui_poll_loop",
    "__version__",
)



# pylint:disable=redefined-outer-name

import json
import sys
from pathlib import Path
from typing import List, Optional, Union, cast

import pytest
from ipykernel.comm import Comm
from ipywidgets import Widget
from pytest_mock import MockerFixture
from traitlets import Unicode

from unfolded.map_sdk.map.widget import SyncWidgetMap
from unfolded.map_sdk.types import WidgetMessage

if sys.version_info >= (3, 8):
    from typing import Literal, TypedDict
else:
    from typing_extensions import Literal, TypedDict

HERE = Path(__file__).parent


class CommMessageDataUpdate(TypedDict):
    method: Literal["update"]
    state: dict
    buffer_paths: list


class CommMessageDataCustom(TypedDict):
    method: Literal["custom"]
    content: WidgetMessage


class CommMessage(TypedDict):
    data: Union[CommMessageDataCustom, CommMessageDataUpdate]
    buffers: Optional[list]


class MockComm(Comm):
    """A mock Comm object.

    Can be used to inspect calls to Comm's open/send/close methods.
    """

    comm_id = Unicode("a-b-c-d")
    kernel = "ipykernel.kernelbase.Kernel"  # type: ignore

    log_open: List[tuple]
    log_send: List[CommMessage]
    log_close: List[tuple]

    def __init__(self, *args, **kwargs):
        self.log_open = []
        self.log_send = []
        self.log_close = []
        super().__init__(*args, **kwargs)

    def open(self, *args, **kwargs):
        self.log_open.append((args, kwargs))

    def send(self, *args, **kwargs):
        if args:
            raise ValueError("only kwargs expected in send comm method")

        self.log_send.append(cast(CommMessage, kwargs))

    def close(self, *args, **kwargs):
        self.log_close.append((args, kwargs))

    # TODO: try to figure out how to mock comm _responses_
    # def handle_msg(self, msg):
    #     """Handle a comm_msg message"""
    #     if self._msg_callback:
    #         self._msg_callback(msg)


@pytest.fixture
def mock_comm(mock_wait_for_future):  # pylint:disable=unused-argument
    _widget_attrs = {}
    undefined = object()

    # Create a single instance of the mocked comm to set as the comm for mock_widget_map
    # and to return from this function to check comm calls
    mock_comm_instance = MockComm()

    _widget_attrs["_ipython_display_"] = Widget._ipython_display_

    def raise_not_implemented(*args, **kwargs):
        raise NotImplementedError()

    Widget._ipython_display_ = raise_not_implemented

    yield mock_comm_instance

    for attr, value in _widget_attrs.items():
        if value is undefined:
            delattr(Widget, attr)
        else:
            setattr(Widget, attr, value)


@pytest.fixture
def mock_wait_for_future(mocker: MockerFixture):
    mocker.patch(
        "unfolded.map_sdk.transport.widget.BlockingWidgetTransport._wait_for_future"
    )


@pytest.fixture
def mock_widget_map(mock_comm: MockComm) -> SyncWidgetMap:
    m = SyncWidgetMap()
    # set mock_comm as the comm for the map
    m.comm = mock_comm
    m.transport._has_loaded = True
    return m


@pytest.fixture
def sentinel2_stac_item():
    path = HERE / "fixtures" / "raster" / "item" / "sentinel-2-l2a.json"
    with open(path) as f:
        return json.load(f)


@pytest.fixture
def sentinel2_stac_collection():
    path = HERE / "fixtures" / "raster" / "collection" / "sentinel-s2-l2a-cogs.json"
    with open(path) as f:
        return json.load(f)



import sys
from concurrent.futures import Future
from time import time
from typing import Callable, Dict, List, Optional, Sequence, Type, Union, overload

from ipywidgets import DOMWidget
from pydantic import parse_obj_as

from unfolded.map_sdk.api.base import Action
from unfolded.map_sdk.api.enums import (
    REVERSE_EVENT_HANDLER_MAP,
    EventType,
    is_event_type,
)
from unfolded.map_sdk.api.event_api import EventHandlers
from unfolded.map_sdk.errors import MapSDKException, UnfoldedStudioException
from unfolded.map_sdk.poll import run_ui_poll_loop
from unfolded.map_sdk.transport.base import BaseInteractiveTransport
from unfolded.map_sdk.types import (
    ErrorResponse,
    EventResponse,
    MessageResponse,
    ResponseClass,
    StoredFuture,
    WidgetMessage,
)
from unfolded.map_sdk.utils.serialization import serialize_action

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class BaseWidgetTransport(BaseInteractiveTransport):
    widget: DOMWidget
    futures: Dict[str, StoredFuture]
    event_handlers: Dict[str, Callable]

    def __init__(
        self,
        widget: DOMWidget,
    ) -> None:
        super().__init__()
        self.widget = widget
        self.futures: Dict[str, StoredFuture] = {}
        self.event_handlers = {}
        self._has_loaded = False

    def set_event_handlers(self, event_handlers: Dict[str, Callable]) -> None:
        self.event_handlers = EventHandlers(
            **{**self.event_handlers, **event_handlers}
        ).dict(exclude_none=True)

    def remove_event_handlers(self, names: Sequence[Union[str, EventType]]) -> None:
        for name in names:
            setattr(self.event_handlers, name, None)

    def _send_widget_message(
        self, content: Union[WidgetMessage, Dict], buffers: Optional[List[bytes]]
    ) -> None:
        """Send message to JS using Jupyter Widgets' messaging protocol"""
        self.widget.send(content=content, buffers=buffers)

    def receive_message(
        self,
        content: Union[EventResponse, MessageResponse, ErrorResponse],
        buffers: Optional[List[bytes]] = None,
    ) -> None:
        """Receive message from JS"""

        if content["type"] == "event":
            self.receive_message_event(content, buffers=buffers)

        elif content["type"] == "error":
            self.receive_message_error(content, buffers=buffers)

        elif content["type"] == "response":
            self.receive_message_valid(content, buffers=buffers)

    def receive_message_event(
        self, content: EventResponse, buffers: Optional[List[bytes]] = None
    ) -> None:
        # pylint: disable=unused-argument
        """Receive event notification from JS"""
        event_type = content.get("eventType")

        if not event_type or not is_event_type(event_type):
            raise MapSDKException(f"{event_type} is not a supported event type")

        event_type = EventType(event_type)
        event_name = REVERSE_EVENT_HANDLER_MAP.get(event_type)
        if event_name == "on_load":
            self._has_loaded = True
        elif event_name:
            callback = self.event_handlers.get(event_name)
            data = content.get("data")

            if callback and data is not None:
                callback(data)

    def receive_message_error(
        self, content: ErrorResponse, buffers: Optional[List[bytes]] = None
    ) -> None:
        # pylint: disable=unused-argument
        """Receive error response message from JS"""
        message_id = content.get("messageId")
        error = content.get("error")

        if not message_id or message_id not in self.futures:
            return

        future_ref = self.futures.pop(message_id)
        future_ref.future.set_exception(UnfoldedStudioException(error))

    def receive_message_valid(
        self, content: MessageResponse, buffers: Optional[List[bytes]] = None
    ) -> None:
        # pylint: disable=unused-argument
        """Receive response message from JS"""
        message_id = content.get("messageId")
        data = content.get("data")

        if not message_id or message_id not in self.futures:
            return

        future_ref = self.futures.pop(message_id)
        response_class = future_ref.response_class

        if response_class is not None:
            try:
                # parse_obj_as should work with any arbitrary pydantic-compatible
                # types
                # https://pydantic-docs.helpmanual.io/usage/models/#parsing-data-into-a-specified-type
                parsed = parse_obj_as(response_class, data)
                future_ref.future.set_result(parsed)
                return
            except Exception as e:
                future_ref.future.set_exception(e)

        future_ref.future.set_result(data)


class BlockingWidgetTransport(BaseWidgetTransport):
    def send_action(
        self,
        *,
        action: Action,
        response_class: Optional[Type[ResponseClass]] = None,
    ) -> Optional[ResponseClass]:
        if not self._has_loaded:
            raise MapSDKException("Cannot call map method before rendering map.")

        content, buffers = serialize_action(action)
        self._send_widget_message(content, buffers)

        future: "Future[ResponseClass]" = Future()
        stored_future: StoredFuture[Type[ResponseClass]] = StoredFuture(
            future=future, response_class=response_class
        )
        self.futures[str(action.message_id)] = stored_future

        result = self._wait_for_future(future, response_optional=True)
        if isinstance(result, Exception):
            raise result

        return result

    def send_action_non_null(
        self,
        *,
        action: Action,
        response_class: Optional[Type[ResponseClass]] = None,
    ) -> ResponseClass:
        if not self._has_loaded:
            raise MapSDKException("Cannot call map method before rendering map.")

        content, buffers = serialize_action(action)
        self._send_widget_message(content, buffers)

        future: "Future[ResponseClass]" = Future()
        stored_future: StoredFuture[Type[ResponseClass]] = StoredFuture(
            future=future, response_class=response_class
        )
        self.futures[str(action.message_id)] = stored_future

        result = self._wait_for_future(future, response_optional=False)
        if isinstance(result, Exception):
            raise result

        return result

    @overload
    def _wait_for_future(
        self,
        future: "Future[ResponseClass]",
        response_optional: Literal[True],
        timeout: float = 5,
    ) -> Optional[ResponseClass]:
        ...

    @overload
    def _wait_for_future(
        self,
        future: "Future[ResponseClass]",
        response_optional: Literal[False],
        timeout: float = 5,
    ) -> ResponseClass:
        ...

    def _wait_for_future(
        self,
        future: "Future[ResponseClass]",
        response_optional: bool,
        timeout: float = 5,
    ) -> Optional[ResponseClass]:
        """Wait in a blocking way for future to be completed

        Args:
            future: Future object to wait on
            response_optional: if True, the response type can be `None`. Otherwise, a `None`
                response from JS will be raised as an exception.
            timeout: Timeout period in seconds after which polling will finish

        Returns:
            Result of future object (could be `None` if response_optional is True). Can also return an Exception stored within the future.
        """
        start_time = time()
        sentinel_object = object()

        def poll_callback():
            """Callback to be run in run_ui_poll_loop

            Polls for completion of future up to specified timeout
            """
            if future.done():
                # Checking for `.exception()` prevents raising the exception here, which would
                # happen if we called `.result()` and an exception had been set
                if future.exception():
                    return future.exception()

                # Returning None from the callback will not break out of the polling
                if future.result() is None:
                    if response_optional:
                        return sentinel_object
                    else:
                        return UnfoldedStudioException(
                            "Got unexpected None response from widget"
                        )

                return future.result()

            if time() - start_time > timeout:
                return UnfoldedStudioException("Timeout waiting for widget response")

            return None

        poll_response = run_ui_poll_loop(poll_callback, sleep=1 / 10, n=4)

        if poll_response is sentinel_object:
            return None

        return poll_response



# type: ignore
"""
Tools for working with async tasks
"""
import asyncio
import threading


class AsyncThread:
    @staticmethod
    def _worker(loop):
        asyncio.set_event_loop(loop)
        loop.run_forever()
        loop.close()

    def __init__(self):
        self._loop = asyncio.new_event_loop()
        self._thread = threading.Thread(target=AsyncThread._worker, args=(self._loop,))
        self._thread.start()

    def terminate(self):
        def _stop(loop):
            loop.stop()

        if self._loop is None:
            return

        self.call_soon(_stop, self._loop)
        self._thread.join()
        self._loop, self._thread = None, None

    def __del__(self):
        self.terminate()

    def submit(self, func, *args, **kwargs):
        """
        Run async func with args/kwargs in separate thread, returns Future object.
        """
        return asyncio.run_coroutine_threadsafe(func(*args, **kwargs), self._loop)

    def wrap(self, func):
        def sync_func(*args, **kwargs):
            return self.submit(func, *args, **kwargs).result()

        return sync_func

    def call_soon(self, func, *args):
        """
        Call normal (non-async) function with arguments in the processing thread
        it's just a wrapper over `loop.call_soon_threadsafe()`

        Returns a handle with `.cancel`, not a full on Future
        """
        return self._loop.call_soon_threadsafe(func, *args)

    @property
    def loop(self):
        return self._loop



"""rename_package

This script is designed to rename a wheel archive so that the package has a different name while the
code layout stays the same. This means that it can be installed under a different name, and pip will
show a different name, but imports remain `import unfolded.map_sdk`.

python rename_package.py --input path_to_wheel.whl --name amazon.map-sdk --output output_dir
"""
import re
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from zipfile import ZipFile

import click


def extract_wheel(origin: Path, destination: Path) -> None:
    with ZipFile(origin) as zf:
        zf.extractall(destination)


def get_dist_info_dir(path: Path) -> Path:
    dist_info_dirs = [d for d in Path(path).iterdir() if "dist-info" in d.name]
    assert len(dist_info_dirs) == 1, "Only one dist-info dir should exist"
    return dist_info_dirs[0]


def get_data_dir(path: Path) -> Path:
    data_dirs = [d for d in Path(path).iterdir() if "data" in d.name]
    assert len(data_dirs) == 1, "Only one data dir should exist"
    return data_dirs[0]


def rename_dir(dir_: Path, new_folder_name: str) -> None:
    new_dir_name = dir_.parent / dir_.name.replace("unfolded.map_sdk", new_folder_name)
    dir_.replace(new_dir_name)


def change_lines_in_record_file(record_path: Path, new_folder_name: str) -> None:
    with open(record_path) as f:
        lines = f.readlines()

    fmt = r"^unfolded.map_sdk-[\d\.]+\.(data|dist-info)/"
    lines = [
        l.replace("unfolded.map_sdk", new_folder_name) if re.match(fmt, l) else l
        for l in lines
    ]

    with open(record_path, "w") as f:
        f.write("".join(lines))


def change_name_in_metadata_file(metadata_path: Path, new_name: str) -> None:
    with open(metadata_path) as f:
        text = f.read()

    text = text.replace("Name: unfolded.map-sdk", f"Name: {new_name}")

    with open(metadata_path, "w") as f:
        f.write(text)


def create_wheel(origin: Path, destination: Path) -> Path:
    zipped_path = Path(shutil.make_archive(str(destination), "zip", origin))
    # Need to rename since shutil.make_archive appends .zip
    return zipped_path.rename(zipped_path.parent / zipped_path.name.rstrip(".zip"))


def rename_package(path: Path, new_name: str, output_dir: Path) -> None:
    new_folder_name = new_name.replace("-", "_")
    output_dir.mkdir(parents=True, exist_ok=True)
    new_wheel_path = output_dir / path.name.replace("unfolded.map_sdk", new_folder_name)

    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        extract_wheel(path, tmp_dir)

        dist_info_dir = get_dist_info_dir(tmp_dir)
        data_dir = get_data_dir(tmp_dir)

        change_name_in_metadata_file(dist_info_dir / "METADATA", new_name)
        change_lines_in_record_file(dist_info_dir / "RECORD", new_folder_name)
        rename_dir(dist_info_dir, new_folder_name)
        rename_dir(data_dir, new_folder_name)

        create_wheel(tmp_dir, new_wheel_path)


@click.command()
@click.option(
    "-i",
    "--input",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
    help="Path to input wheel.",
    required=True,
)
@click.option(
    "-n",
    "--name",
    type=str,
    help="New package name.",
    default="amzn.map-sdk",
    show_default=True,
)
@click.option(
    "-o",
    "--output",
    type=click.Path(dir_okay=True, file_okay=False, writable=True),
    required=True,
    help="Path to output directory.",
)
def main(input: str, name: str, output: str):  # pylint:disable=redefined-builtin
    rename_package(Path(input), name, Path(output))


if __name__ == "__main__":
    main()



import pytest

from tests.conftest import MockComm
from unfolded.map_sdk.errors import MapSDKException
from unfolded.map_sdk.map.widget import SyncWidgetMap


class TestWidgetTransport:
    def test_event_handler_merge(self):

        m = SyncWidgetMap()
        transport = m.transport

        func1 = lambda view: print(1)
        func2 = lambda mouse_event: print(2)
        func3 = lambda view: print(3)
        func4 = lambda filter: print(4)

        event_handlers_1 = {"on_view_update": func1, "on_click": func2}

        transport.set_event_handlers(event_handlers_1)
        assert transport.event_handlers == event_handlers_1

        event_handlers_2 = {"on_view_update": func3, "on_filter_update": func4}

        transport.set_event_handlers(event_handlers_2)

        event_handlers_final = {
            "on_view_update": func3,
            "on_click": func2,
            "on_filter_update": func4,
        }
        assert transport.event_handlers == event_handlers_final

    def test_calling_method_before_rendering_raises(
        self, mock_comm: MockComm  # pylint:disable=unused-argument
    ):
        m = SyncWidgetMap()
        with pytest.raises(MapSDKException):
            m.set_view(longitude=0)



from tests._utils import check_expected_comm_message
from tests.conftest import MockComm
from unfolded.map_sdk.map.widget import SyncWidgetMap

DATASET_UUID = "0e939627-8ea7-4db6-8c92-3dd943166a01"
FILTER_UUID = "e02b67ea-20d4-4613-808e-9a8fdbd81e6f"


class TestGetFilters:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_filters()

        expected = {
            "type": "v1/map-sdk-get-filters",
            "args": [],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestGetFilterById:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_filter_by_id(FILTER_UUID)

        expected = {
            "type": "v1/map-sdk-get-filter-by-id",
            "args": [FILTER_UUID],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestAddFilter:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-add-filter",
            "args": [
                {
                    "type": "range",
                    "sources": [{"dataId": DATASET_UUID, "fieldName": "test"}],
                    "value": [0, 100],
                }
            ],
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        filter_ = {
            "type": "range",
            "sources": [{"data_id": DATASET_UUID, "field_name": "test"}],
            "value": (0, 100),
        }
        mock_widget_map.add_filter(filter_)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        filter_ = {
            "type": "range",
            "sources": [{"data_id": DATASET_UUID, "field_name": "test"}],
            "value": (0, 100),
        }
        mock_widget_map.add_filter(**filter_)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)


class TestUpdateFilter:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-update-filter",
            "args": [
                FILTER_UUID,
                {
                    "type": "range",
                    "value": [0, 50],
                    "sources": [{"dataId": DATASET_UUID, "fieldName": "test-2"}],
                },
            ],
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        values = {
            "value": (0, 50),
            "sources": [{"data_id": DATASET_UUID, "field_name": "test-2"}],
        }
        mock_widget_map.update_filter(FILTER_UUID, values)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        values = {
            "value": (0, 50),
            "sources": [{"data_id": DATASET_UUID, "field_name": "test-2"}],
        }
        mock_widget_map.update_filter(FILTER_UUID, **values)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)


class TestRemoveFilter:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.remove_filter(FILTER_UUID)

        expected = {
            "type": "v1/map-sdk-remove-filter",
            "args": [FILTER_UUID],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestUpdateTimeline:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-update-timeline",
            "args": [FILTER_UUID, {"view": "side", "isAnimating": True}],
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        values = {"view": "side", "is_animating": True}
        mock_widget_map.update_timeline(FILTER_UUID, values)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        values = {"view": "side", "is_animating": True}
        mock_widget_map.update_timeline(FILTER_UUID, **values)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)



DUMMY_MESSAGE_UUID = "ede3fafc-4945-41b7-ac64-16faa896a47a"
LAYER_UUID = "997c21eb-604b-49df-add9-b2fcc615a243"

LOCAL_RESPONSE = {
    "id": "2853b4cf-005f-47f5-a6f0-54d275e77d9c",
    "label": "earthquakes",
    "color": [0, 92, 255],
    "fields": [
        {
            "name": "DateTime",
            "label": "DateTime",
            "type": "timestamp",
            "timeFormat": "YYYY/M/D HH:mm:ss.SSSS",
        },
        {"name": "Latitude", "label": "Latitude", "type": "real"},
        {"name": "MagType", "label": "MagType", "type": "string"},
        {"name": "NbStations", "label": "NbStations", "type": "integer"},
    ],
    "type": "local",
}

VECTOR_TILE_RESPONSE = {
    "id": "56a78675-5729-449e-8961-1b0f30562f4f",
    "label": "Test",
    "color": [143, 47, 191],
    "fields": [
        {"name": "AGE_21_64", "label": "AGE_21_64", "type": "real"},
    ],
    "type": "vector-tile",
    "metadata": {
        "dataUrl": "https://storage.googleapis.com/unfolded_public/vector-tile/cb_v2",
        "metadataUrl": "https://storage.googleapis.com/unfolded_public/vector-tile/cb_v2/metadata.json",
        "id": "56a78675-5729-449e-8961-1b0f30562f4f",
        "format": "",
        "label": "Test",
        "metaJson": {
            "vector_layers": [
                {
                    "id": "ny_age_s20",
                    "description": "",
                    "minzoom": 7,
                    "maxzoom": 10,
                    "fields": {
                        "AGE_21_64": "Number",
                    },
                },
            ],
            "tilestats": {
                "layerCount": 53,
                "layers": [
                    {
                        "layer": "ny_age_s20",
                        "count": 241486,
                        "geometry": "Polygon",
                        "attributeCount": 4,
                        "attributes": [
                            {
                                "attribute": "AGE_21_64",
                                "count": 1000,
                                "type": "number",
                                "values": [
                                    0,
                                    0.002849002849002849,
                                    0.0034602076124567475,
                                    0.006289308176100629,
                                    0.007861635220125786,
                                    0.008,
                                    0.046822742474916385,
                                    0.048314606741573035,
                                ],
                                "min": 0,
                                "max": 1,
                            },
                        ],
                    },
                ],
            },
        },
        "bounds": [-176.684659, 17.926837, 174.144841, 71.334586],
        "center": [-74.003906, 40.846927, 10],
        "maxZoom": 10,
        "minZoom": 0,
        "fields": [
            {
                "name": "AGE_21_64",
                "id": "AGE_21_64",
                "format": "",
                "filterProps": {
                    "fieldType": "real",
                    "domain": [0, 1],
                    "histogram": None,
                    "value": [0, 1],
                    "type": "range",
                    "typeOptions": ["range"],
                    "gpu": True,
                    "step": 0.001,
                },
                "type": "real",
                "analyzerType": "FLOAT",
            },
        ],
        "name": "_DEL_POP/ALL",
        "description": "_DEL_POP/ALL",
    },
}

RASTER_TILE_RESPONSE = {
    "id": "f65ce936-cc3c-4de2-9914-290c93d786e6",
    "label": "Test",
    "color": [192, 108, 132],
    "fields": [],
    "type": "raster-tile",
    "metadata": {
        "metadataUrl": "https://raw.githubusercontent.com/UnfoldedInc/examples/master/raster/planet-skysat-opendata.json",
        "id": "20201213_045438_ss01_u0001",
        "format": "",
        "label": "Test",
        "type": "Feature",
        "stac_version": "1.0.0",
        "stac_extensions": [
            "https://stac-extensions.github.io/eo/v1.0.0/schema.json",
            "https://stac-extensions.github.io/raster/v1.0.0/schema.json",
            "https://stac-extensions.github.io/view/v1.0.0/schema.json",
        ],
        "properties": {
            "providers": [
                {
                    "name": "Planet",
                    "description": "Contact Planet at [planet.com/contact-sales](https://www.planet.com/contact-sales/)",
                    "roles": ["producer", "processor"],
                    "url": "http://planet.com",
                }
            ],
            "gsd": 1.01,
            "created": "2020-12-13T09:20:24Z",
            "updated": "2020-12-13T09:20:24Z",
            "constellation": "skysat",
            "platform": "SS01",
            "eo:cloud_cover": 0,
            "view:off_nadir": 25.5,
            "view:sun_azimuth": 159.8,
            "view:sun_elevation": 42.4,
            "pl:clear_confidence_percent": 59,
            "pl:clear_percent": 28,
            "pl:cloud_percent": 0,
            "pl:ground_control_ratio": 0.95,
            "pl:heavy_haze_percent": 0,
            "pl:item_type": "SkySatCollect",
            "pl:light_haze_percent": 72,
            "pl:pixel_resolution": 0.5,
            "pl:publishing_stage": "finalized",
            "pl:quality_category": "test",
            "pl:satellite_azimuth": 141.7,
            "pl:shadow_percent": 0,
            "pl:snow_ice_percent": 0,
            "pl:strip_id": "s1_20201213T045438Z",
            "pl:visible_confidence_percent": 59,
            "pl:visible_percent": 100,
            "datetime": "2020-12-13T04:54:38.312000Z",
        },
        "geometry": {
            "coordinates": [
                [
                    [88.80278308162045, 21.713415693279156],
                    [88.80313138501904, 21.71337310737007],
                    [88.80317550773762, 21.700562568435377],
                ]
            ],
            "type": "Polygon",
        },
        "links": [
            {"rel": "root", "href": "../collection.json", "type": "application/json"},
            {
                "rel": "collection",
                "href": "../collection.json",
                "type": "application/json",
            },
            {"rel": "parent", "href": "../collection.json", "type": "application/json"},
        ],
        "assets": {
            "visual:ortho_visual": {
                "href": "https://storage.googleapis.com/open-cogs/planet-stac/20201213_045438_ss01_u0001/20201213_045438_ss01_u0001_visual_file_format.tif",
                "type": "image/tiff; application=geotiff; profile=cloud-optimized",
                "roles": ["data"],
                "eo:bands": [
                    {"name": "Blue", "common_name": "blue"},
                ],
                "raster:bands": [
                    {"data_type": "uint8"},
                ],
            },
            "metadata": {
                "href": "https://storage.googleapis.com/open-cogs/planet-stac/20201213_045438_ss01_u0001/20201213_045438_ss01_u0001_metadata.json",
                "type": "application/json",
                "roles": ["metadata"],
            },
        },
        "bbox": [
            88.7123238494617,
            21.700562568435377,
            88.80317550773762,
            21.89915173899179,
        ],
        "collection": "planet-stac-skysat",
    },
}



from tests._utils import check_expected_comm_message
from tests.conftest import MockComm
from unfolded.map_sdk.map.widget import SyncWidgetMap

LAYER_UUID = "997c21eb-604b-49df-add9-b2fcc615a243"


class TestGetView:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_view(0)

        expected = {
            "type": "v1/map-sdk-get-view",
            "args": [0],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestSetView:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-set-view",
            "args": [
                {
                    "longitude": 8.650367,
                    "latitude": 47.271057,
                    "zoom": 5.0,
                }
            ],
            "options": {"index": 0},
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        view = {"latitude": 47.271057, "longitude": 8.650367, "zoom": 5}
        mock_widget_map.set_view(view, 0)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        view = {"latitude": 47.271057, "longitude": 8.650367, "zoom": 5}
        mock_widget_map.set_view(**view, index=0)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)


class TestGetViewLimits:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_view_limits(0)

        expected = {
            "type": "v1/map-sdk-get-view-limits",
            "args": [0],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestSetViewLimits:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-set-view-limits",
            "args": [
                {
                    "minZoom": 3,
                    "maxZoom": 8,
                }
            ],
            "options": {"index": 0},
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        view_limits = {"min_zoom": 3, "max_zoom": 8}
        mock_widget_map.set_view_limits(view_limits, 0)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        view_limits = {"min_zoom": 3, "max_zoom": 8}
        mock_widget_map.set_view_limits(**view_limits, index=0)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)


class TestGetMapControlVisibility:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_map_control_visibility()

        expected = {
            "type": "v1/map-sdk-get-map-control-visibility",
            "args": [],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestSetMapControlVisibility:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-set-map-control-visibility",
            "args": [
                {
                    "legend": False,
                    "toggle-3d": True,
                }
            ],
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        visibility = {"legend": False, "toggle_3d": True}
        mock_widget_map.set_map_control_visibility(visibility)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        visibility = {"legend": False, "toggle_3d": True}
        mock_widget_map.set_map_control_visibility(**visibility)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)


class TestGetSplitMode:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_split_mode()

        expected = {
            "type": "v1/map-sdk-get-split-mode",
            "args": [],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestSetSplitMode:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-set-split-mode",
            "args": [
                "swipe",
                {
                    "layers": [[LAYER_UUID], []],
                    "isViewSynced": True,
                    "isZoomSynced": True,
                },
            ],
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        options = {
            "layers": [[LAYER_UUID], []],
            "is_view_synced": True,
            "is_zoom_synced": True,
        }
        mock_widget_map.set_split_mode("swipe", options)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        options = {
            "layers": [[LAYER_UUID], []],
            "is_view_synced": True,
            "is_zoom_synced": True,
        }
        mock_widget_map.set_split_mode("swipe", **options)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)


class TestSetTheme:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.set_theme(preset="light", background_color="blue")

        expected = {
            "type": "v1/map-sdk-set-theme",
            "args": [{"preset": "light", "options": {"backgroundColor": "blue"}}],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestGetMapConfig:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_map_config()

        expected = {
            "type": "v1/map-sdk-get-map-config",
            "args": [],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestSetMapConfig:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        config = {"map": "config"}
        dataset = {"id": "dataset-id", "type": "local", "data": "dataset-data"}

        mock_widget_map.set_map_config(config, additional_datasets=[dataset])

        expected = {
            "type": "v1/map-sdk-set-map-config",
            "args": [
                {"map": "config"},
                {
                    "additionalDatasets": [
                        {"id": "dataset-id", "type": "local", "data": "dataset-data"}
                    ]
                },
            ],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestGetMapStyles:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_map_styles()

        expected = {
            "type": "v1/map-sdk-get-map-styles",
            "args": [],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)



from typing import Callable

import ipykernel

from unfolded.map_sdk._poll import run_ui_poll_loop as run_ui_poll_loop_v4
from unfolded.map_sdk._poll_v56 import run_ui_poll_loop as run_ui_poll_loop_v56

__all__ = "run_ui_poll_loop"


def get_polling_fn() -> Callable:
    """Return valid implementation of jupyter-ui-poll for ipykernel version

    Returns:
        Callable: implementation of run_ui_poll_loop for this ipykernel version
    """
    ipykernel_major_version = int(ipykernel.__version__[0])

    if ipykernel_major_version < 5:
        # NOTE: haven't tested this function on ipykernel v3 and prior
        return run_ui_poll_loop_v4
    else:
        return run_ui_poll_loop_v56


run_ui_poll_loop = get_polling_fn()



"""Helpers to deduce what type of Python environment is currently being used
"""

import os
from enum import Enum
from typing import Union

import IPython

__all__ = ("Environment", "CURRENT_ENVIRONMENT", "default_height")


class Environment(str, Enum):
    AZURE = "azure"
    COCALC = "cocalc"
    COLAB = "colab"
    CONSOLE = "console"
    DATABRICKS = "databricks"
    JUPYTER_LAB = "jupyterlab"
    JUPYTER_NOTEBOOK = "notebook"
    KAGGLE = "kaggle"
    NTERACT = "nteract"
    VSCODE = "vscode"


def deduce_current_environment() -> Environment:
    """Deduce current Python environment"""
    # Ported from Plotly under the MIT license:
    # https://github.com/plotly/plotly.py/blob/2c2dd6ab2eeff73c782457f33c590c1d09a97625/packages/python/plotly/plotly/io/_renderers.py#L455-L536

    if IPython and IPython.get_ipython():
        # Try to detect environment so that we can enable a useful
        # default renderer
        try:
            import google.colab  # pylint:disable=import-outside-toplevel,unused-import

            return Environment.COLAB
        except ImportError:
            pass

        # Check if we're running in a Kaggle notebook
        if os.path.exists("/kaggle/input"):
            return Environment.KAGGLE

        # Check if we're running in an Azure Notebook
        if "AZURE_NOTEBOOKS_HOST" in os.environ:
            return Environment.AZURE

        # Check if we're running in VSCode
        if "VSCODE_PID" in os.environ:
            return Environment.VSCODE

        # Check if we're running in nteract
        if "NTERACT_EXE" in os.environ:
            return Environment.NTERACT

        # Check if we're running in CoCalc
        if "COCALC_PROJECT_ID" in os.environ:
            return Environment.COCALC

        if "DATABRICKS_RUNTIME_VERSION" in os.environ:
            return Environment.DATABRICKS

        # Check if we're running in IPython terminal
        if IPython.get_ipython().__class__.__name__ == "TerminalInteractiveShell":
            return Environment.CONSOLE

    # Fall back to console environment
    return Environment.CONSOLE


CURRENT_ENVIRONMENT = deduce_current_environment()


def default_height() -> Union[str, int]:
    """Provide default widget heights for current environment"""
    height: Union[str, int] = "100%"

    if CURRENT_ENVIRONMENT in [
        Environment.COLAB,
        Environment.VSCODE,
        Environment.DATABRICKS,
    ]:
        # Colab: 100% height shows up as very small
        # VSCode: 100% height shows up as very small
        # Databricks: 100% height leads to the iframe height
        # growing infinitely in Databricks due to their resizing logic
        height = 500

    return height



from pydantic import parse_obj_as

from unfolded.map_sdk.api.dataset_api import (
    Dataset,
    LocalDataset,
    RasterTileDataset,
    RasterTileLocalItemMetadata,
    VectorTileDataset,
    VectorTileLocalMetadata,
)
from unfolded.map_sdk.api.enums import ActionType
from unfolded.map_sdk.utils.serialization import FUNCTION_MAPPING

from .fixtures.dataset_api import (
    LOCAL_RESPONSE,
    RASTER_TILE_RESPONSE,
    VECTOR_TILE_RESPONSE,
)


class TestDataset:
    """Tests relating to dataset serialization/deserialization"""

    def test_deserialize(self):

        local_dataset = parse_obj_as(Dataset, LOCAL_RESPONSE)
        assert isinstance(local_dataset, LocalDataset)

        vector_tile_dataset = parse_obj_as(Dataset, VECTOR_TILE_RESPONSE)
        assert isinstance(vector_tile_dataset, VectorTileDataset)
        assert isinstance(vector_tile_dataset.metadata, VectorTileLocalMetadata)

        raster_tile_dataset = parse_obj_as(Dataset, RASTER_TILE_RESPONSE)
        assert isinstance(raster_tile_dataset, RasterTileDataset)
        assert isinstance(raster_tile_dataset.metadata, RasterTileLocalItemMetadata)


class TestFunctionMapping:
    """Tests relating to the function mapping we maintain for the HTML map"""

    def test_function_mapping(self):

        for key in ActionType:
            # Event handler actions are not translated to Javascript
            if key not in [
                ActionType.SET_EVENT_HANDLERS,
                ActionType.REMOVE_EVENT_HANDLERS,
            ]:
                assert key in FUNCTION_MAPPING



from typing import Any, Dict, List, Optional, Union

from pydantic import StrictBool, StrictStr

from unfolded.map_sdk.api.base import Action, CamelCaseBaseModel, Number
from unfolded.map_sdk.api.enums import ActionType, EffectType
from unfolded.map_sdk.transport.base import (
    BaseInteractiveTransport,
    BaseNonInteractiveTransport,
    BaseTransport,
)
from unfolded.map_sdk.utils.validators import validate_kwargs


class _EffectUpdateProps(CamelCaseBaseModel):
    is_enabled: Optional[StrictBool]
    """Flag indicating whether effect is enabled or not."""

    parameters: Optional[
        Dict[StrictStr, Union[Number, List[Number], StrictBool, StrictStr]]
    ]
    """Dictionary that maps fields that the effect uses."""


class _EffectCreationProps(_EffectUpdateProps):
    id: Optional[StrictStr]
    """Unique identifier of the effect."""

    type: Optional[EffectType]
    """Type of the effect."""

    is_enabled: Optional[StrictBool]
    """Flag indicating whether effect is enabled or not."""

    parameters: Optional[
        Dict[StrictStr, Union[Number, List[Number], StrictBool, StrictStr]]
    ]
    """Dictionary that maps fields that the effect uses."""


class Effect(_EffectCreationProps):
    """Type encapsulating effect properties."""

    id: StrictStr
    """Unique identifier of the effect."""

    type: EffectType
    """Type of the effect."""

    is_enabled: StrictBool
    """Flag indicating whether effect is enabled or not."""

    parameters: Optional[
        Dict[StrictStr, Union[Number, List[Number], StrictBool, StrictStr]]
    ]
    """Dictionary that maps fields that the effect uses."""


###########
# ACTIONS #
###########


class GetEffectsAction(Action):
    type: ActionType = ActionType.GET_EFFECTS


class GetEffectByIdAction(Action):
    class Meta(Action.Meta):
        args = ["effect_id"]

    type: ActionType = ActionType.GET_EFFECT_BY_ID
    effect_id: StrictStr


class AddEffectAction(Action):
    class Meta(Action.Meta):
        args = ["effect"]

    type: ActionType = ActionType.ADD_EFFECT
    effect: _EffectCreationProps


class UpdateEffectAction(Action):
    class Meta(Action.Meta):
        args = ["effect_id", "values"]

    type: ActionType = ActionType.UPDATE_EFFECT
    effect_id: StrictStr
    values: _EffectUpdateProps


class RemoveEffectAction(Action):
    class Meta(Action.Meta):
        args = ["effect_id"]

    type: ActionType = ActionType.REMOVE_EFFECT
    effect_id: StrictStr


###########
# METHODS #
###########


class BaseEffectApiMethods:
    transport: BaseTransport

    @validate_kwargs(positional_only=["effect"])
    def add_effect(
        self, effect: Union[_EffectCreationProps, dict, None] = None, **kwargs: Any
    ) -> Optional[Effect]:
        """Adds a new effect to the map.

        Args:
            effect (Union[EffectCreationProps, dict, None], optional): The effect to add. Defaults to None.

        Returns (widget map only):
            Effect: The effect that was added.
        """
        action = AddEffectAction(effect=effect if effect is not None else kwargs)
        return self.transport.send_action_non_null(action=action, response_class=Effect)

    @validate_kwargs(positional_only=["effect_id", "values"])
    def update_effect(
        self,
        effect_id: str,
        values: Union[_EffectUpdateProps, dict, None] = None,
        **kwargs: Any
    ) -> Optional[Effect]:
        """Updates an existing effect with given values.

        Args:
            effect_id (str): The id of the effect to update.
            values (Union[EffectUpdateProps, dict, None]): The values to update.

        Returns (widget map only)
            Effect: The updated effect.
        """
        action = UpdateEffectAction(
            effect_id=effect_id, values=values if values is not None else kwargs
        )
        return self.transport.send_action_non_null(action=action, response_class=Effect)

    @validate_kwargs(positional_only=["effect_id"])
    def remove_effect(self, effect_id: str) -> None:
        """Removes a effect from the map.

        Args:
            effect_id (str): The id of the effect to remove

        Returns:
            None
        """
        action = RemoveEffectAction(effect_id=effect_id)
        self.transport.send_action(action=action, response_class=None)


class BaseInteractiveEffectApiMethods:
    transport: BaseInteractiveTransport

    def get_effects(self) -> List[Effect]:
        """Gets all the effects currently available in the map.

        Returns:
            List[Effect]: An array of effects.
        """
        action = GetEffectsAction()
        return self.transport.send_action_non_null(
            action=action, response_class=List[Effect]
        )

    @validate_kwargs(positional_only=["effect_id"])
    def get_effect_by_id(self, effect_id: str) -> Optional[Effect]:
        """Retrieves a effect by its identifier if it exists.

        Args:
            effect_id (str): Identifier of the effect to get.

        Returns:
            Optional[effect]: Effect with a given identifier, or null if one doesn't exist.
        """
        action = GetEffectByIdAction(effect_id=effect_id)
        return self.transport.send_action(action=action, response_class=Effect)


class EffectApiNonInteractiveMixin(BaseEffectApiMethods):
    """Effect methods that are supported in non-interactive (i.e. pure HTML) maps"""

    transport: BaseNonInteractiveTransport

    def add_effect(
        self, effect: Union[_EffectCreationProps, dict, None] = None, **kwargs: Any
    ) -> None:
        super().add_effect(effect, **kwargs)
        return

    @validate_kwargs(positional_only=["effect_id", "values"])
    def update_effect(
        self, effect_id: str, values: Union[_EffectUpdateProps, dict], **kwargs: Any
    ) -> None:
        super().update_effect(effect_id, values, **kwargs)
        return


class EffectApiInteractiveMixin(BaseEffectApiMethods, BaseInteractiveEffectApiMethods):
    """Effect methods that are supported in interactive (i.e. widget) maps"""

    transport: BaseInteractiveTransport

    def add_effect(
        self, effect: Union[_EffectCreationProps, dict, None] = None, **kwargs: Any
    ) -> Effect:
        return super().add_effect(effect, **kwargs)

    @validate_kwargs(positional_only=["effect_id", "values"])
    def update_effect(
        self,
        effect_id: str,
        values: Union[_EffectUpdateProps, dict, None] = None,
        **kwargs: Any
    ) -> Effect:
        return super().update_effect(effect_id, values, **kwargs)



from typing import Dict

from unfolded.map_sdk.api.enums import ActionType

# Mapping from message type to MapApi function name for message dispatch
FUNCTION_MAPPING: Dict[ActionType, str] = {
    # Map API
    ActionType.GET_VIEW: "getView",
    ActionType.SET_VIEW: "setView",
    ActionType.GET_VIEW_LIMITS: "getViewLimits",
    ActionType.SET_VIEW_LIMITS: "setViewLimits",
    ActionType.GET_VIEW_MODE: "getViewMode",
    ActionType.SET_VIEW_MODE: "setViewMode",
    ActionType.GET_MAP_CONTROL_VISIBILITY: "getMapControlVisibility",
    ActionType.SET_MAP_CONTROL_VISIBILITY: "setMapControlVisibility",
    ActionType.GET_SPLIT_MODE: "getSplitMode",
    ActionType.SET_SPLIT_MODE: "setSplitMode",
    ActionType.GET_MAP_CONFIG: "getMapConfig",
    ActionType.SET_MAP_CONFIG: "setMapConfig",
    ActionType.GET_MAP_STYLES: "getMapStyles",
    ActionType.SET_THEME: "setUiTheme",
    # Filter API
    ActionType.GET_FILTERS: "getFilters",
    ActionType.GET_FILTER_BY_ID: "getFilterById",
    ActionType.ADD_FILTER: "addFilter",
    ActionType.UPDATE_FILTER: "updateFilter",
    ActionType.REMOVE_FILTER: "removeFilter",
    ActionType.UPDATE_TIMELINE: "updateTimeline",
    # Datasets API
    ActionType.GET_DATASETS: "getDatasets",
    ActionType.GET_DATASET_BY_ID: "getDatasetById",
    ActionType.ADD_DATASET: "addDataset",
    ActionType.ADD_TILE_DATASET: "addTileDataset",
    ActionType.UPDATE_DATASET: "updateDataset",
    ActionType.REMOVE_DATASET: "removeDataset",
    ActionType.REPLACE_DATASET: "replaceDataset",
    ActionType.GET_DATASET_WITH_DATA: "getDatasetWithData",
    # Layer API
    ActionType.GET_LAYERS: "getLayers",
    ActionType.GET_LAYER_BY_ID: "getLayerById",
    ActionType.ADD_LAYER: "addLayer",
    ActionType.UPDATE_LAYER: "updateLayer",
    ActionType.REMOVE_LAYER: "removeLayer",
    ActionType.GET_LAYER_GROUPS: "getLayerGroups",
    ActionType.GET_LAYER_GROUP_BY_ID: "getLayerGroupById",
    ActionType.ADD_LAYER_GROUP: "addLayerGroup",
    ActionType.UPDATE_LAYER_GROUP: "updateLayerGroup",
    ActionType.REMOVE_LAYER_GROUP: "removeLayerGroup",
    ActionType.GET_LAYER_TIMELINE: "getLayerTimeline",
    ActionType.UPDATE_LAYER_TIMELINE: "updateLayerTimeline",
    # Effect API
    ActionType.GET_EFFECTS: "getEffects",
    ActionType.GET_EFFECT_BY_ID: "getEffectById",
    ActionType.ADD_EFFECT: "addEffect",
    ActionType.UPDATE_EFFECT: "updateEffect",
    ActionType.REMOVE_EFFECT: "removeEffect",
}



import sys
from typing import Any, Callable, Dict, List, Optional, Union

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from pydantic import StrictBool, StrictStr

from unfolded.map_sdk.api.base import Action, CamelCaseBaseModel, Number, TimeRange
from unfolded.map_sdk.api.enums import ActionType, LayerType
from unfolded.map_sdk.transport.base import (
    BaseInteractiveTransport,
    BaseNonInteractiveTransport,
    BaseTransport,
)
from unfolded.map_sdk.utils.validators import validate_kwargs


class VisualChannel(CamelCaseBaseModel):
    name: StrictStr
    type: StrictStr


VisualChannels = Dict[StrictStr, Union[VisualChannel, StrictStr, None]]


class _PartialLayerConfig(CamelCaseBaseModel):
    visual_channels: Optional[VisualChannels]
    """Dictionary of visualization properties that the layer supports."""

    vis_config: Optional[Dict[StrictStr, Any]]
    """General layer settings."""

    column_mode: Optional[
        Literal["points", "geojson", "table", "neighbors", "LAT_LNG", "H3"]
    ]
    """Column mode for the layer (some of the layers support multiple input data formats)."""

    text_label: Optional[Dict[StrictStr, Any]]
    """Label settings."""


class LayerConfig(_PartialLayerConfig):
    visual_channels: VisualChannels
    """Dictionary of visualization properties that the layer supports."""

    vis_config: Dict[StrictStr, Any]
    """General layer settings."""

    column_mode: Optional[
        Literal["points", "geojson", "table", "neighbors", "LAT_LNG", "H3"]
    ]
    """Column mode for the layer (some of the layers support multiple input data formats)."""

    text_label: Optional[Dict[StrictStr, Any]]
    """Label settings."""


class _LayerUpdateProps(CamelCaseBaseModel):
    type: Optional[LayerType]
    """Type of the layer."""

    data_id: Optional[StrictStr]
    """Unique identifier of the dataset this layer visualizes."""

    fields: Optional[Dict[StrictStr, Optional[StrictStr]]]
    """Dictionary that maps fields that the layer requires for visualization to appropriate dataset fields."""

    label: Optional[StrictStr]
    """Canonical label of this layer."""

    is_visible: Optional[StrictBool]
    """Flag indicating whether layer is visible or not."""

    config: Optional[_PartialLayerConfig]
    """Layer configuration specific to its type."""


class _LayerCreationProps(_LayerUpdateProps):
    id: Optional[StrictStr]
    """Unique identifier of the layer."""

    type: Optional[LayerType]
    """Type of the layer."""

    data_id: StrictStr
    """Unique identifier of the dataset this layer visualizes."""

    fields: Optional[Dict[StrictStr, Optional[StrictStr]]]
    """Dictionary that maps fields that the layer requires for visualization to appropriate dataset fields."""

    label: Optional[StrictStr]
    """Canonical label of this layer."""

    is_visible: Optional[StrictBool]
    """Flag indicating whether layer is visible or not."""

    config: Optional[_PartialLayerConfig]
    """Layer configuration specific to its type."""


class Layer(_LayerCreationProps):
    """Type encapsulating layer properties."""

    id: StrictStr
    """Unique identifier of the layer."""

    type: Optional[LayerType]
    """Type of the layer."""

    data_id: StrictStr
    """Unique identifier of the dataset this layer visualizes."""

    fields: Dict[StrictStr, Optional[StrictStr]]
    """Dictionary that maps fields that the layer requires for visualization to appropriate dataset fields."""

    label: StrictStr
    """Canonical label of this layer."""

    is_visible: StrictBool
    """Flag indicating whether layer is visible or not."""

    config: LayerConfig
    """Layer configuration specific to its type."""


class _LayerGroupUpdateProps(CamelCaseBaseModel):
    label: Optional[StrictStr]
    """Canonical label of this layer group."""

    is_visible: Optional[StrictBool]
    """Flag indicating whether layer group is visible or not."""

    layer_ids: Optional[List[StrictStr]]
    """Layers that are part of this group, sorted in the order in which they are shown."""


class _LayerGroupCreationProps(_LayerGroupUpdateProps):
    id: Optional[StrictStr]
    """Unique identifier of the layer group."""

    label: Optional[StrictStr]
    """Canonical label of this layer group."""

    is_visible: Optional[StrictBool]
    """Flag indicating whether layer group is visible or not."""

    layer_ids: Optional[List[StrictStr]]
    """Layers that are part of this group, sorted in the order in which they are shown."""


class LayerGroup(CamelCaseBaseModel):
    """A conceptual group used to organize layers."""

    id: StrictStr
    """Unique identifier of the layer group."""

    label: StrictStr
    """Canonical label of this layer group."""

    is_visible: StrictBool
    """Flag indicating whether layer group is visible or not."""

    layer_ids: List[StrictStr]
    """Layers that are part of this group, sorted in the order in which they are shown."""


class _LayerTimelineUpdateProps(CamelCaseBaseModel):
    current_time: Optional[Number]
    """Current time on the timeline in milliseconds."""

    is_animating: Optional[StrictBool]
    """Flag indicating whether the timeline is animating or not."""

    is_visible: Optional[StrictBool]
    """Flag indicating whether the timeline is visible or not."""

    animation_speed: Optional[Number]
    """Speed at which timeline is animating."""

    time_format: Optional[StrictStr]
    """Time format that the timeline is using in day.js supported format.

    https://day.js.org/docs/en/display/format
    """

    timezone: Optional[StrictStr]
    """Timezone that the timeline is using in tz format.

    https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    """


class LayerTimeline(_LayerTimelineUpdateProps):
    current_time: Number
    """Current time on the timeline in milliseconds."""

    domain: TimeRange
    """Range of time that the timeline shows."""

    is_animating: StrictBool
    """Flag indicating whether the timeline is animating or not."""

    is_visible: StrictBool
    """Flag indicating whether the timeline is visible or not."""

    animation_speed: Number
    """Speed at which timeline is animating."""

    time_format: StrictStr
    """Time format that the timeline is using in day.js supported format.

    https://day.js.org/docs/en/display/format
    """

    timezone: Optional[StrictStr]
    """Timezone that the timeline is using in tz format.

    https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    """

    time_steps: Optional[List[Number]]
    """Step duration for all the animation keyframes in milliseconds."""


class LayerEventHandlers(CamelCaseBaseModel):
    on_layer_timeline_update: Optional[Callable[[LayerTimeline], None]]


###########
# ACTIONS #
###########


class GetLayersAction(Action):
    type: ActionType = ActionType.GET_LAYERS


class GetLayerByIdAction(Action):
    class Meta(Action.Meta):
        args = ["layer_id"]

    type: ActionType = ActionType.GET_LAYER_BY_ID
    layer_id: StrictStr


class AddLayerAction(Action):
    class Meta(Action.Meta):
        args = ["layer"]

    type: ActionType = ActionType.ADD_LAYER
    layer: _LayerCreationProps


class UpdateLayerAction(Action):
    class Meta(Action.Meta):
        args = ["layer_id", "values"]

    type: ActionType = ActionType.UPDATE_LAYER
    layer_id: StrictStr
    values: _LayerUpdateProps


class RemoveLayerAction(Action):
    class Meta(Action.Meta):
        args = ["layer_id"]

    type: ActionType = ActionType.REMOVE_LAYER
    layer_id: StrictStr


class GetLayerGroupsAction(Action):
    type: ActionType = ActionType.GET_LAYER_GROUPS


class GetLayerGroupByIdAction(Action):
    class Meta(Action.Meta):
        args = ["layer_group_id"]

    type: ActionType = ActionType.GET_LAYER_GROUP_BY_ID
    layer_group_id: StrictStr


class AddLayerGroupAction(Action):
    class Meta(Action.Meta):
        args = ["layer_group"]

    type: ActionType = ActionType.ADD_LAYER_GROUP
    layer_group: _LayerGroupCreationProps


class UpdateLayerGroupAction(Action):
    class Meta(Action.Meta):
        args = ["layer_group_id", "values"]

    type: ActionType = ActionType.UPDATE_LAYER_GROUP
    layer_group_id: StrictStr
    values: _LayerGroupUpdateProps


class RemoveLayerGroupAction(Action):
    class Meta(Action.Meta):
        args = ["layer_group_id"]

    type: ActionType = ActionType.REMOVE_LAYER_GROUP
    layer_group_id: StrictStr


class GetLayerTimelineAction(Action):
    type: ActionType = ActionType.GET_LAYER_TIMELINE


class UpdateLayerTimelineAction(Action):
    class Meta(Action.Meta):
        args = ["values"]

    type: ActionType = ActionType.UPDATE_LAYER_TIMELINE
    values: _LayerTimelineUpdateProps


###########
# METHODS #
###########


class BaseLayerApiMethods:
    transport: BaseTransport

    @validate_kwargs(positional_only=["layer"])
    def add_layer(
        self, layer: Union[_LayerCreationProps, dict, None] = None, **kwargs: Any
    ) -> Optional[Layer]:
        """Adds a new layer to the map.

        Args:
            layer (Union[LayerCreationProps, dict, None], optional): The layer to add. Defaults to None.

        Returns (widget map only):
            Layer: The layer that was added.
        """
        action = AddLayerAction(layer=layer if layer is not None else kwargs)
        return self.transport.send_action_non_null(action=action, response_class=Layer)

    @validate_kwargs(positional_only=["layer_id", "values"])
    def update_layer(
        self,
        layer_id: str,
        values: Union[_LayerUpdateProps, dict, None] = None,
        **kwargs: Any
    ) -> Optional[Layer]:
        """Updates an existing layer with given values.

        Args:
            layer_id (str): The id of the layer to update.
            values (Union[LayerUpdateProps, dict, None]): The values to update.

        Returns (widget map only)
            Layer: The updated layer.
        """
        action = UpdateLayerAction(
            layer_id=layer_id, values=values if values is not None else kwargs
        )
        return self.transport.send_action_non_null(action=action, response_class=Layer)

    @validate_kwargs(positional_only=["layer_id"])
    def remove_layer(self, layer_id: str) -> None:
        """Removes a layer from the map.

        Args:
            layer_id (str): The id of the layer to remove

        Returns:
            None
        """
        action = RemoveLayerAction(layer_id=layer_id)
        self.transport.send_action(action=action, response_class=None)

    @validate_kwargs(positional_only=["layer_group"])
    def add_layer_group(
        self,
        layer_group: Union[_LayerGroupCreationProps, dict, None] = None,
        **kwargs: Any
    ) -> Optional[LayerGroup]:
        """Adds a new layer group to the map.

        Args:
            layer_group (Union[_LayerGroupCreationProps, dict, None], optional): The layer group to add. Defaults to None.

        Returns (widget map only):
            LayerGroup: The layer group that was added.
        """
        action = AddLayerGroupAction(
            layer_group=layer_group if layer_group is not None else kwargs
        )
        return self.transport.send_action_non_null(
            action=action, response_class=LayerGroup
        )

    @validate_kwargs(positional_only=["layer_group_id", "values"])
    def update_layer_group(
        self,
        layer_group_id: str,
        values: Union[_LayerGroupUpdateProps, dict, None] = None,
        **kwargs: Any
    ) -> Optional[LayerGroup]:
        """Updates an existing layer group with given values.

        Args:
            layer_group_id (str): The id of the layer group to update.
            values (Union[_LayerGroupUpdateProps, dict, None]): The values to update.

        Returns (widget map only):
            LayerGroup: The updated layer group.
        """
        action = UpdateLayerGroupAction(
            layer_group_id=layer_group_id,
            values=values if values is not None else kwargs,
        )
        return self.transport.send_action_non_null(
            action=action, response_class=LayerGroup
        )

    @validate_kwargs(positional_only=["layer_group_id"])
    def remove_layer_group(self, layer_group_id: str) -> None:
        """Removes a layer group from the map.

        Args:
            layer_group_id (str): The id of the layer group to remove

        Returns:
            None
        """
        action = RemoveLayerGroupAction(layer_group_id=layer_group_id)
        self.transport.send_action(action=action, response_class=None)

    @validate_kwargs(positional_only=["values"])
    def update_layer_timeline(
        self, values: Union[_LayerTimelineUpdateProps, dict, None] = None, **kwargs: Any
    ) -> Optional[LayerTimeline]:
        """Updates the current layer timeline configuration.

        Args:
            values (Union[LayerTimelineUpdateProps, dict, None], optional): The new layer timeline values. Defaults to None.

        Returns (widget map only):
            LayerTimeline: The updated layer timeline.
        """
        action = UpdateLayerTimelineAction(
            values=values if values is not None else kwargs
        )
        return self.transport.send_action_non_null(
            action=action, response_class=LayerTimeline
        )


class BaseInteractiveLayerApiMethods:
    transport: BaseInteractiveTransport

    def get_layers(self) -> List[Layer]:
        """Gets all the layers currently available in the map.

        Returns:
            List[Layer]: An array of layers.
        """
        action = GetLayersAction()
        return self.transport.send_action_non_null(
            action=action, response_class=List[Layer]
        )

    @validate_kwargs(positional_only=["layer_id"])
    def get_layer_by_id(self, layer_id: str) -> Optional[Layer]:
        """Retrieves a layer by its identifier if it exists.

        Args:
            layer_id (str): Identifier of the layer to get.

        Returns:
            Optional[Layer]: Layer with a given identifier, or null if one doesn't exist.
        """
        action = GetLayerByIdAction(layer_id=layer_id)
        return self.transport.send_action(action=action, response_class=Layer)

    def get_layer_groups(self) -> List[LayerGroup]:
        """Gets all the layer groups currently available in the map.

        Returns:
            List[LayerGroup]: An array of layer groups.
        """
        action = GetLayerGroupsAction()
        return self.transport.send_action_non_null(
            action=action, response_class=List[LayerGroup]
        )

    def get_layer_group_by_id(self, layer_group_id: str) -> Optional[LayerGroup]:
        """Retrieves a layer group by its identifier if it exists.

        Args:
            layer_group_id (str): Identifier of the layer group to get.

        Returns:
            Optional[LayerGroup]: Layer group with a given identifier, or null if one doesn't exist.
        """
        action = GetLayerGroupByIdAction(layer_group_id=layer_group_id)
        return self.transport.send_action(action=action, response_class=LayerGroup)

    def get_layer_timeline(self) -> Optional[LayerTimeline]:
        """Gets the current layer timeline configuration.

        Returns:
            Optional[LayerTimeline]: The layer timeline configuration, or null if it doesn't exist.
        """
        action = GetLayerTimelineAction()
        return self.transport.send_action(action=action, response_class=LayerTimeline)


class LayerApiNonInteractiveMixin(BaseLayerApiMethods):
    """Layer methods that are supported in non-interactive (i.e. pure HTML) maps"""

    transport: BaseNonInteractiveTransport

    def add_layer(
        self, layer: Union[_LayerCreationProps, dict, None] = None, **kwargs: Any
    ) -> None:
        super().add_layer(layer, **kwargs)
        return

    @validate_kwargs(positional_only=["layer_id", "values"])
    def update_layer(
        self,
        layer_id: str,
        values: Union[_LayerUpdateProps, dict, None] = None,
        **kwargs: Any
    ) -> None:
        super().update_layer(layer_id, values, **kwargs)
        return

    @validate_kwargs(positional_only=["layer_group"])
    def add_layer_group(
        self,
        layer_group: Union[_LayerGroupCreationProps, dict, None] = None,
        **kwargs: Any
    ) -> None:
        super().add_layer_group(layer_group, **kwargs)
        return

    @validate_kwargs(positional_only=["layer_group_id", "values"])
    def update_layer_group(
        self,
        layer_group_id: str,
        values: Union[_LayerGroupUpdateProps, dict, None] = None,
        **kwargs: Any
    ) -> None:
        super().update_layer_group(layer_group_id, values, **kwargs)
        return

    @validate_kwargs(positional_only=["values"])
    def update_layer_timeline(
        self, values: Union[_LayerTimelineUpdateProps, dict, None] = None, **kwargs: Any
    ) -> None:
        super().update_layer_timeline(values, **kwargs)
        return


class LayerApiInteractiveMixin(BaseLayerApiMethods, BaseInteractiveLayerApiMethods):
    """Layer methods that are supported in interactive (i.e. widget) maps"""

    transport: BaseInteractiveTransport

    def add_layer(
        self, layer: Union[_LayerCreationProps, dict, None] = None, **kwargs: Any
    ) -> Layer:
        return super().add_layer(layer, **kwargs)

    @validate_kwargs(positional_only=["layer_id", "values"])
    def update_layer(
        self,
        layer_id: str,
        values: Union[_LayerUpdateProps, dict, None] = None,
        **kwargs: Any
    ) -> Layer:
        return super().update_layer(layer_id, values, **kwargs)

    @validate_kwargs(positional_only=["layer_group"])
    def add_layer_group(
        self,
        layer_group: Union[_LayerGroupCreationProps, dict, None] = None,
        **kwargs: Any
    ) -> LayerGroup:
        return super().add_layer_group(layer_group, **kwargs)

    @validate_kwargs(positional_only=["layer_group_id", "values"])
    def update_layer_group(
        self,
        layer_group_id: str,
        values: Union[_LayerGroupUpdateProps, dict, None] = None,
        **kwargs: Any
    ) -> LayerGroup:
        return super().update_layer_group(layer_group_id, values, **kwargs)

    @validate_kwargs(positional_only=["values"])
    def update_layer_timeline(
        self, values: Union[_LayerTimelineUpdateProps, dict, None] = None, **kwargs: Any
    ) -> LayerTimeline:
        return super().update_layer_timeline(values, **kwargs)



import inspect
import sys
from inspect import signature
from typing import Union

import pytest

from unfolded.map_sdk.api.dataset_api import (
    BaseDatasetApiMethods,
    DatasetApiInteractiveMixin,
    DatasetApiNonInteractiveMixin,
)
from unfolded.map_sdk.api.effect_api import (
    BaseEffectApiMethods,
    EffectApiInteractiveMixin,
    EffectApiNonInteractiveMixin,
)
from unfolded.map_sdk.api.filter_api import (
    BaseFilterApiMethods,
    FilterApiInteractiveMixin,
    FilterApiNonInteractiveMixin,
)
from unfolded.map_sdk.api.layer_api import (
    BaseLayerApiMethods,
    LayerApiInteractiveMixin,
    LayerApiNonInteractiveMixin,
)
from unfolded.map_sdk.api.map_api import (
    BaseMapApiMethods,
    MapApiInteractiveMixin,
    MapApiNonInteractiveMixin,
)

if sys.version_info >= (3, 8):
    from typing import get_args, get_origin
else:
    from typing_extensions import get_args, get_origin

API_CLASSES = [
    [BaseMapApiMethods, MapApiInteractiveMixin, MapApiNonInteractiveMixin],
    [BaseDatasetApiMethods, DatasetApiInteractiveMixin, DatasetApiNonInteractiveMixin],
    [BaseLayerApiMethods, LayerApiInteractiveMixin, LayerApiNonInteractiveMixin],
    [BaseFilterApiMethods, FilterApiInteractiveMixin, FilterApiNonInteractiveMixin],
    [BaseEffectApiMethods, EffectApiInteractiveMixin, EffectApiNonInteractiveMixin],
]


class TestSubclass:
    """Tests relating to Map SDK API subclasses.
    Specifically, whether interactive and non-interactive mixins
    have the right type signatures on their methods
    """

    @pytest.mark.parametrize("base,interactive,non_interactive", API_CLASSES)
    def test_subclasses(self, base, interactive, non_interactive):

        base_methods = inspect.getmembers(base, predicate=inspect.isfunction)
        interactive_methods = dict(
            inspect.getmembers(interactive, predicate=inspect.isfunction)
        )
        non_interactive_methods = dict(
            inspect.getmembers(non_interactive, predicate=inspect.isfunction)
        )

        for func_name, func in base_methods:
            sig = signature(func)
            base_return_type = sig.return_annotation
            # Match Optional return type
            if get_origin(base_return_type) is Union and get_args(base_return_type)[
                -1
            ] is type(None):

                *declared_return_type, _ = get_args(base_return_type)

                sub_return_type = signature(
                    interactive_methods[func_name]
                ).return_annotation

                # Check if subclass return type is Union
                if get_origin(sub_return_type) is Union:
                    sub_return_type = list(get_args(sub_return_type))
                # Otherwise we should only have one non-None type in the Union
                else:
                    assert (
                        len(declared_return_type) == 1
                    ), f"{func_name} in {base.__name__} has too many non-None return types: {declared_return_type}, {func_name} in {interactive.__name__} should return a Union"
                    declared_return_type = declared_return_type[0]

                # Make sure method is subclassed properly
                assert (
                    sub_return_type == declared_return_type
                ), f"{func_name} in {interactive.__name__} has wrong return type"
                assert (
                    signature(non_interactive_methods[func_name]).return_annotation
                    is None
                ), f"{func_name} in {non_interactive.__name__} has non-None return type"

            # match None return type
            elif base_return_type is None:
                if func_name in interactive_methods:
                    sub_func = interactive_methods[func_name]
                    # Check if subclass redefines function
                    assert (
                        sub_func.__qualname__ == func.__qualname__
                    ), f"{func_name} does not need to appear in {interactive.__name__} since return type is None"
                if func_name in non_interactive_methods:
                    sub_func = non_interactive_methods[func_name]
                    # Check if subclass redefines function
                    assert (
                        sub_func.__qualname__ == func.__qualname__
                    ), f"{func_name} does not need to appear in {non_interactive.__name__} since return type is None"
            else:
                raise TypeError(
                    f"{func_name} in {base.__name__} has non-Optional, non-None return type"
                )



from tests._utils import check_expected_comm_message
from tests.conftest import MockComm
from unfolded.map_sdk.map.widget import SyncWidgetMap


class TestGetEffects:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_effects()

        expected = {
            "type": "v1/map-sdk-get-effects",
            "args": [],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestGetEffectById:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_effect_by_id("effect-id")

        expected = {
            "type": "v1/map-sdk-get-effect-by-id",
            "args": ["effect-id"],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestAddEffect:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-add-effect",
            "args": [
                {
                    "id": "effect-id",
                    "type": "ink",
                    "isEnabled": False,
                    "parameters": {"strength": 0.33},
                }
            ],
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        effect = {
            "id": "effect-id",
            "type": "ink",
            "is_enabled": False,
            "parameters": {"strength": 0.33},
        }

        mock_widget_map.add_effect(effect)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        effect = {
            "id": "effect-id",
            "type": "ink",
            "is_enabled": False,
            "parameters": {"strength": 0.33},
        }

        mock_widget_map.add_effect(**effect)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)


class TestUpdateEffect:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-update-effect",
            "args": [
                "effect-id",
                {"parameters": {"strength": 0.22}},
            ],
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        values = {"parameters": {"strength": 0.22}}
        mock_widget_map.update_effect("effect-id", values)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        values = {"parameters": {"strength": 0.22}}
        mock_widget_map.update_effect("effect-id", **values)

        assert check_expected_comm_message(self.expected, mock_comm.log_send)


class TestRemoveEffect:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.remove_effect("effect-id")

        expected = {
            "type": "v1/map-sdk-remove-effect",
            "args": ["effect-id"],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)



import sys
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, overload

from geojson_pydantic import Feature, Polygon
from pydantic import AnyHttpUrl, StrictBool, StrictInt, StrictStr

from unfolded.map_sdk.api.base import (
    Action,
    CamelCaseBaseModel,
    KebabCaseBaseModel,
    Number,
)
from unfolded.map_sdk.api.dataset_api import _DatasetCreationProps
from unfolded.map_sdk.api.enums import ActionType
from unfolded.map_sdk.transport.base import (
    BaseInteractiveTransport,
    BaseNonInteractiveTransport,
    BaseTransport,
)
from unfolded.map_sdk.utils.validators import validate_kwargs

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


ViewMode = Literal["2d", "3d", "globe"]


class _PartialView(CamelCaseBaseModel):
    """Partial version of View for data input type"""

    longitude: Optional[Number]
    """Longitude of the view center [-180, 180]."""

    latitude: Optional[Number]
    """Latitude of the view center [-90, 90]."""

    zoom: Optional[Number]
    """View zoom level [0-22]."""

    pitch: Optional[Number]
    """View pitch value [0-90]."""

    bearing: Optional[Number]
    """View bearing [0-360]."""

    drag_rotate: Optional[StrictBool]
    """Enable rotating with pointer drag."""


class View(_PartialView):
    """Strict version of _PartialView for method return type"""

    longitude: Number
    """Longitude of the view center [-180, 180]."""

    latitude: Number
    """Latitude of the view center [-90, 90]."""

    zoom: Number
    """View zoom level [0-22]."""

    pitch: Number
    """View pitch value [0-90]."""

    bearing: Number
    """View bearing [0-360]."""

    drag_rotate: StrictBool
    """Enable rotating with pointer drag."""


class Bounds(CamelCaseBaseModel):
    min_longitude: Number
    """The west bound."""

    max_longitude: Number
    """The east bound."""

    min_latitude: Number
    """The south bound."""

    max_latitude: Number
    """The north bound."""


class _PartialViewLimits(CamelCaseBaseModel):
    """Partial version of ViewLimits for data input type"""

    min_zoom: Optional[Number]
    """Minimum zoom of the map [0-22]"""

    max_zoom: Optional[Number]
    """Maximum zoom of the map [0-22]"""

    max_bounds: Optional[Bounds]
    """Maximum bounds of the map."""


class ViewLimits(_PartialViewLimits):
    """Strict version of _PartialViewLimits for method return type"""

    min_zoom: Number
    """Minimum zoom of the map [0-22]"""

    max_zoom: Number
    """Maximum zoom of the map [0-22]"""

    max_bounds: Bounds
    """Maximum bounds of the map."""


class _PartialMapControlVisibility(KebabCaseBaseModel):
    """Partial version of View for data input type"""

    legend: Optional[StrictBool]
    """Whether the legend is visible."""

    toggle_3d: Optional[StrictBool]
    """Whether the 3D toggle is visible."""

    split_map: Optional[StrictBool]
    """Whether the split map button is visible."""

    map_draw: Optional[StrictBool]
    """Whether the map draw button is visible."""


class MapControlVisibility(_PartialMapControlVisibility):
    """Strict version of _PartialMapControlVisibility for method return type"""

    legend: StrictBool
    """Whether the legend is visible."""

    toggle_3d: StrictBool
    """Whether the 3D toggle is visible."""

    split_map: StrictBool
    """Whether the split map button is visible."""

    map_draw: StrictBool
    """Whether the map draw button is visible."""


SplitMode = Literal["single", "dual", "swipe"]


class _PartialSplitModeContext(CamelCaseBaseModel):
    layers: Optional[List[List[StrictStr]]]
    """An array of layer ids to show on either side of the split. Only applicable for multi-view split modes."""

    is_view_synced: Optional[StrictBool]
    """Boolean indicating whether views are synced. Only applicable to dual split mode."""

    is_zoom_synced: Optional[StrictBool]
    """Boolean indicating whether zoom is synced between views. Only applicable to dual split mode."""


class SplitModeContext(_PartialSplitModeContext):
    """Base type that contains context around the split map mode."""

    layers: List[List[StrictStr]]
    """An array of layer ids to show on either side of the split. Only applicable for multi-view split modes."""

    is_view_synced: StrictBool
    """Boolean indicating whether views are synced. Only applicable to dual split mode."""

    is_zoom_synced: StrictBool
    """Boolean indicating whether zoom is synced between views. Only applicable to dual split mode."""


class SplitModeDetails(SplitModeContext):
    """Container type for split map mode details."""

    split_mode: SplitMode


SetSplitModeOptions = _PartialSplitModeContext

ThemePresets = Literal["light", "dark"]


class ThemeOptions(CamelCaseBaseModel):

    background_color: Optional[StrictStr]
    """Background color of UI elements"""


class ThemeUpdateProps(CamelCaseBaseModel):

    preset: Optional[ThemePresets]
    """Preset UI theme name"""

    options: Optional[ThemeOptions]
    """UI theme update options"""


class MapStyleLayerGroup(CamelCaseBaseModel):
    """A grouping concept for basemap layers, allowing for interaction on a set of layers."""

    label: StrictStr
    """Layer group text representation.
    The currently supported labels are: label, road, border, building, water, land, 3d building.
    """

    default_visibility: StrictBool
    """Whether map style this group should be visible by default."""


class MapStyleLayerGroupCreationProps(MapStyleLayerGroup):
    """A set of properties required when specifying a map style layer group."""

    filter: StrictStr
    """
    Layer filtering regular expression that determines whether layer with a given id should be part of this group.

    Supports regular JS RegExp syntax (https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/RegExp).
    """


class MapStyleCreationProps(CamelCaseBaseModel):
    """A set of properties required when specifying a map style."""

    id: Optional[StrictStr]
    """Unique identifier of the map style."""

    label: StrictStr
    """User facing map style label."""

    url: AnyHttpUrl
    """URL of the map style descriptor json in Mapbox style format."""

    thumbnail_url: Optional[StrictStr]
    """URL of the thumbnail to show alongside the map style."""

    layer_groups: Optional[List[MapStyleLayerGroupCreationProps]]
    """An array of map style layer groups."""


class MapStyle(MapStyleCreationProps):
    """
    Style applied to the underlying map style. The map style descriptor must conform
    to the Mapbox style specification (https://docs.mapbox.com/mapbox-gl-js/style-spec/).
    """

    id: StrictStr
    """Unique identifier of the map style."""

    label: StrictStr
    """User facing map style label."""

    url: AnyHttpUrl
    """URL of the map style descriptor json in Mapbox style format."""

    thumbnail_url: StrictStr
    """URL of the thumbnail to show alongside the map style."""

    layer_groups: List[MapStyleLayerGroup]  # type:ignore[assignment]
    """An array of map style layer groups."""


class SetMapConfigOptions(CamelCaseBaseModel):

    additional_datasets: Optional[List[_DatasetCreationProps]]
    """Datasets to add to the map before loading the map configuration."""


class PickInfo(CamelCaseBaseModel):
    """Additional context on map data associated to this event."""

    layer_id: StrictStr
    """Identifier of the layer that the mouse event is related to."""

    row_index: Number
    """Index of the row in the data table that picked layer visualizes."""

    row_data: List[Any]
    """Row data that the picked layer visualizes."""


class MouseEvent(CamelCaseBaseModel):
    """Mouse event capturing information such as position and any map features related to it."""

    coordinate: Tuple[Number, Number]
    """Geospatial coordinates [longitude, latitude]."""

    position: Tuple[Number, Number]
    """Mouse position relative to the viewport [x, y]."""

    pick_info: Optional[PickInfo]
    """Additional context on map data associated to this event."""


class GeometrySelectionEvent(CamelCaseBaseModel):
    features: List[Feature[Polygon, Any]]


OnHoverHandlerType = Callable[[MouseEvent], None]
OnClickHandlerType = Callable[[MouseEvent], None]
OnViewUpdateType = Callable[[View], None]
OnGeometrySelectionType = Callable[[GeometrySelectionEvent], None]


class MapEventHandlers(CamelCaseBaseModel):
    """Event handlers which can be registered for receiving notifications of certain map events."""

    on_click: Optional[OnClickHandlerType]
    """Called whenever user clicks the visible part of the map."""

    on_hover: Optional[OnHoverHandlerType]
    """Called whenever a mouse hovers over a visible part of the map."""

    on_view_update: Optional[OnViewUpdateType]
    """Called whenever visible map viewport changes."""

    on_geometry_selection: Optional[OnGeometrySelectionType]
    """Called whenever a geometry is drawn in the map."""


###########
# ACTIONS #
###########


class GetViewAction(Action):
    """Action payload sent with `get_view` calls"""

    class Meta(Action.Meta):
        args = ["index"]

    type: ActionType = ActionType.GET_VIEW
    index: StrictInt


class SetViewAction(Action):
    """Action payload sent with `set_view` calls"""

    class Meta(Action.Meta):
        args = ["view"]
        options = ["index"]

    type: ActionType = ActionType.SET_VIEW
    view: _PartialView
    index: StrictInt
    """Index of the view to update, relevant when multi-view split mode is used and views are unlocked."""


class GetViewLimitsAction(Action):
    """Action payload sent with `get_view_limits` calls"""

    class Meta(Action.Meta):
        args = ["index"]

    type: ActionType = ActionType.GET_VIEW_LIMITS
    index: StrictInt


class SetViewLimitsAction(Action):
    """Action payload sent with `set_view_limits` calls"""

    class Meta(Action.Meta):
        args = ["view_limits"]
        options = ["index"]

    type: ActionType = ActionType.SET_VIEW_LIMITS
    view_limits: _PartialViewLimits
    index: StrictInt
    """Index of the view limits to update, relevant when multi-view split mode is used and views are unlocked."""


class GetViewModeAction(Action):
    """Action payload sent with `get_view_mode` calls"""

    type: ActionType = ActionType.GET_VIEW_MODE


class SetViewModeAction(Action):
    """Action payload sent with `set_view_mode` calls"""

    class Meta(Action.Meta):
        args = ["view_mode"]

    type: ActionType = ActionType.SET_VIEW_MODE
    view_mode: ViewMode


class GetMapControlVisibilityAction(Action):
    """Action payload sent with `get_map_control_visibility` calls"""

    type: ActionType = ActionType.GET_MAP_CONTROL_VISIBILITY


class SetMapControlVisibilityAction(Action):
    """Action payload sent with `set_map_control_visibility` calls"""

    class Meta(Action.Meta):
        args = ["visibility"]

    type: ActionType = ActionType.SET_MAP_CONTROL_VISIBILITY
    visibility: _PartialMapControlVisibility


class GetSplitModeAction(Action):
    """Action payload sent with `get_split_mode` calls"""

    type: ActionType = ActionType.GET_SPLIT_MODE


class SetSplitModeAction(Action):
    """Action payload sent with `set_split_mode` calls"""

    class Meta(Action.Meta):
        args = ["split_mode", "options"]

    type: ActionType = ActionType.SET_SPLIT_MODE
    split_mode: SplitMode
    options: SetSplitModeOptions


class SetThemeAction(Action):
    """Action payload sent with `set_theme`calls"""

    class Meta(Action.Meta):
        args = ["theme"]

    type: ActionType = ActionType.SET_THEME
    theme: ThemeUpdateProps


class GetMapStylesAction(Action):

    type: ActionType = ActionType.GET_MAP_STYLES


class GetMapConfigAction(Action):

    type: ActionType = ActionType.GET_MAP_CONFIG


class SetMapConfigAction(Action):
    class Meta(Action.Meta):
        args = ["config", "options"]

    type: ActionType = ActionType.SET_MAP_CONFIG

    config: Dict[StrictStr, Any]
    options: Optional[SetMapConfigOptions]


###########
# METHODS #
###########


class BaseMapApiMethods:

    transport: BaseTransport

    @validate_kwargs(positional_only=["view"])
    def set_view(
        self, view: Union[View, dict, None] = None, index: int = 0, **kwargs: Number
    ) -> Optional[View]:
        """Sets a new view state for the map.

        Args:
            view: View model instance or a dict with the same attributes, all optional.

        Kwargs (cannot be used with positional arg):
            latitude: Longitude of the view center [-180, 180].
            longitude: Latitude of the view center [-90, 90].
            zoom: View zoom level [0-22].
            pitch: View pitch value [0-90].
            bearing: View bearing [0-360].

        Returns (widget map only):
            View: The new view state of the map.
        """
        action = SetViewAction(view=view if view is not None else kwargs, index=index)
        return self.transport.send_action_non_null(action=action, response_class=View)

    @validate_kwargs(positional_only=["view_limits"])
    def set_view_limits(
        self,
        view_limits: Union[ViewLimits, dict, None] = None,
        index: int = 0,
        **kwargs: Union[Number, Bounds, dict],
    ) -> Optional[ViewLimits]:
        """Sets new view limits for the map.

        Args:
            view_limits: ViewLimits model instance or a dict with the same attributes, all optional.

        Kwargs (cannot be used with positional arg):
            min_zoom: Minimum zoom of the map [0-22].
            max_zoom: Maximum zoom of the map [0-22].
            max_bounds: a Bounds object or a dict with the keys (`min_longitude`, `max_longitude`, `min_latitude`, `max_latitude`).

        Returns (widget map only):
            ViewLimits: The new view limits of the map.
        """
        action = SetViewLimitsAction(
            view_limits=view_limits if view_limits is not None else kwargs, index=index
        )
        return self.transport.send_action_non_null(
            action=action, response_class=ViewLimits
        )

    @validate_kwargs(positional_only=["view_mode"])
    def set_view_mode(
        self,
        view_mode: ViewMode,
    ) -> Optional[ViewMode]:
        """Sets map view mode.

        Args:
            view_mode: ViewMode for the map

        Returns (widget map only):
            ViewMode: The new view mode of the map.
        """
        action = SetViewModeAction(view_mode=view_mode)
        return self.transport.send_action_non_null(
            # mypy has problems with a parse_obj_as() and union types, so we
            # chose to ignore. See: https://github.com/python/mypy/issues/15354
            action=action,
            response_class=Optional[ViewMode],  # type: ignore
        )

    @validate_kwargs(positional_only=["visibility"])
    def set_map_control_visibility(
        self, visibility: Union[MapControlVisibility, dict, None] = None, **kwargs: bool
    ) -> Optional[MapControlVisibility]:
        """Sets new map control visibility for the map.

        Args:
            visibility: MapControlVisibility model instance or a dict with the same attributes, all optional.

        Kwargs (cannot be used with positional arg):
            legend: Whether the legend is visible.
            toggle_3d: Whether the 3D toggle is visible.
            split_map: Whether the split map button is visible.
            map_draw: Whether the map draw button is visible.

        Returns (widget map only):
            MapControlVisibility: The new map control visibility of the map.
        """
        action = SetMapControlVisibilityAction(
            visibility=visibility if visibility is not None else kwargs
        )
        return self.transport.send_action_non_null(
            action=action, response_class=MapControlVisibility
        )

    @validate_kwargs(positional_only=["split_mode", "options"])
    def set_split_mode(
        self,
        split_mode: SplitMode,
        options: Union[SplitModeContext, dict, None] = None,
        **kwargs: Any,
    ) -> Optional[SplitModeDetails]:
        """Sets a new split mode for the map.

        Args:
            split_mode: The desired split mode.
            options: SplitModeContext model instance or a dict with the same attributes.

        Kwargs (cannot be used with positional arg):
            layers: Arrays with layer Ids per each post-split map section.

        Returns (widget map only):
            SplitModeDetails: The new split mode of the map.
        """
        action = SetSplitModeAction(
            split_mode=split_mode, options=options if options is not None else kwargs
        )
        return self.transport.send_action_non_null(
            action=action, response_class=SplitModeDetails
        )

    def set_theme(
        self,
        *,
        preset: Optional[ThemePresets] = None,
        background_color: Optional[str] = None,
    ) -> None:
        """Sets a UI theme for the map

        Kwargs:
            preset: a preset theme: "light" or "dark"
            background_color: background color of UI elements

        Returns:
            None
        """
        props = ThemeUpdateProps(
            preset=preset, options={"background_color": background_color}
        )
        action = SetThemeAction(theme=props)
        return self.transport.send_action(action=action, response_class=None)

    @validate_kwargs(positional_only=["config", "options"])
    def set_map_config(
        self,
        config: dict,
        options: Union[dict, SetMapConfigOptions, None] = None,
        **kwargs,
    ) -> None:
        """Loads the given configuration into the current map.

        Args:
            config (dict): Configuration to load into the map. For details around the format see https://location.foursquare.com/developer/docs/studio-map-configuration.
            options (Optional[SetMapConfigOptions]): A set of options for the map configuration.

        Kwargs (cannot be used with positional arg):
            additional_datasets: Datasets to add to the map before loading the map configuration.

        Returns:
            None
        """
        action = SetMapConfigAction(
            config=config, options=options if options else kwargs
        )
        return self.transport.send_action(action=action, response_class=None)


class BaseInteractiveMapApiMethods:

    transport: BaseInteractiveTransport

    @overload
    def get_view(self, index: Literal[0] = 0) -> View:
        ...

    @overload
    def get_view(self, index: int = 0) -> Optional[View]:
        ...

    def get_view(self, index: int = 0) -> Optional[View]:
        """Gets the current view state of the map.

        Returns:
            The current view state of the map.
        """
        action = GetViewAction(index=index)
        return self.transport.send_action(action=action, response_class=View)

    @overload
    def get_view_limits(self, index: Literal[0] = 0) -> ViewLimits:
        ...

    @overload
    def get_view_limits(self, index: int = 0) -> Optional[ViewLimits]:
        ...

    def get_view_limits(self, index: int = 0) -> Optional[ViewLimits]:
        """Gets the current view limits of the map.

        Returns:
            The current view limits of the map.
        """
        action = GetViewLimitsAction(index=index)
        return self.transport.send_action(action=action, response_class=ViewLimits)

    def get_view_mode(self) -> ViewMode:
        """Gets the current view mode of the map.

        Returns:
            The current view mode of the map.
        """
        action = GetViewModeAction()
        return self.transport.send_action_non_null(
            # mypy has problems with a parse_obj_as() and union types, so we
            # chose to ignore. See: https://github.com/python/mypy/issues/15354
            action=action,
            response_class=ViewMode,  # type: ignore
        )

    def get_map_styles(self) -> List[MapStyle]:
        """Gets the currently available map styles.

        Returns:
            List[MapStyle]: All currently available map styles.
        """
        action = GetMapStylesAction()
        return self.transport.send_action_non_null(
            action=action, response_class=List[MapStyle]
        )

    def get_map_control_visibility(self) -> MapControlVisibility:
        """Gets the current map control visibility of the map.

        Returns:
            The current map control visibility of the map.
        """
        action = GetMapControlVisibilityAction()
        return self.transport.send_action_non_null(
            action=action, response_class=MapControlVisibility
        )

    def get_split_mode(self) -> SplitModeContext:
        """Gets the current split mode of the map.

        Returns:
            The current split mode of the map.
        """
        action = GetSplitModeAction()
        return self.transport.send_action_non_null(
            action=action, response_class=SplitModeContext
        )

    def get_map_config(self) -> dict:
        """Gets the configuration representing the current map state.

        Returns:
            dict: Current map configuration. For details around the format see @link https://location.foursquare.com/developer/docs/studio-map-configuration.
        """
        action = GetMapConfigAction()
        return self.transport.send_action_non_null(action=action, response_class=dict)


class MapApiNonInteractiveMixin(BaseMapApiMethods):
    """Map methods that are supported in non-interactive (i.e. pure HTML) maps"""

    transport: BaseNonInteractiveTransport

    def set_view(
        self, view: Union[View, dict, None] = None, index: int = 0, **kwargs: Number
    ) -> None:
        super().set_view(view, index=index, **kwargs)
        return

    def set_view_limits(
        self,
        view_limits: Union[ViewLimits, dict, None] = None,
        index: int = 0,
        **kwargs: Union[Number, Bounds, dict],
    ) -> None:
        super().set_view_limits(view_limits, index=index, **kwargs)
        return

    def set_view_mode(self, view_mode: ViewMode) -> None:
        super().set_view_mode(view_mode)
        return

    def set_map_control_visibility(
        self, visibility: Union[MapControlVisibility, dict, None] = None, **kwargs: bool
    ) -> None:
        super().set_map_control_visibility(visibility, **kwargs)
        return

    def set_split_mode(
        self,
        split_mode: SplitMode,
        options: Union[SplitModeContext, dict, None] = None,
        **kwargs: Any,
    ) -> None:
        super().set_split_mode(split_mode, options, **kwargs)
        return


class MapApiInteractiveMixin(BaseMapApiMethods, BaseInteractiveMapApiMethods):
    """Map methods that are supported in interactive (i.e. widget) maps"""

    transport: BaseInteractiveTransport

    def set_view(
        self, view: Union[View, dict, None] = None, index: int = 0, **kwargs: Number
    ) -> View:
        return super().set_view(view, index=index, **kwargs)

    def set_view_limits(
        self,
        view_limits: Union[ViewLimits, dict, None] = None,
        index: int = 0,
        **kwargs: Union[Number, Bounds, dict],
    ) -> ViewLimits:
        return super().set_view_limits(view_limits, index=index, **kwargs)

    def set_view_mode(self, view_mode: ViewMode) -> ViewMode:
        return super().set_view_mode(view_mode)

    def set_map_control_visibility(
        self, visibility: Union[MapControlVisibility, dict, None] = None, **kwargs: bool
    ) -> MapControlVisibility:
        return super().set_map_control_visibility(visibility, **kwargs)

    def set_split_mode(
        self,
        split_mode: SplitMode,
        options: Union[SplitModeContext, dict, None] = None,
        **kwargs: Any,
    ) -> SplitModeDetails:
        return super().set_split_mode(split_mode, options, **kwargs)



"""Vendored jupyter-ui-poll version 0.1.3 under the MIT license

We use jupyter-ui-poll to offer synchronous messaging between Python and JS. As of November 2021,
Colab uses an older ipykernel version 4.x. Version 0.1.3 of jupyter-ui-poll supported ipykernel 4.x
but newer versions support only ipykernel 5.x and 6.x. In
https://github.com/Kirill888/jupyter-ui-poll/issues/18 the author was reluctant to support ipykernel
4.x going forward. To solve this, we vendor here version 0.1.3 of jupyter-ui-poll, then can switch
between polling implementation depending on the version of ipykernel.
"""
import asyncio
import sys
import time
from contextlib import contextmanager
from typing import Any, Callable, Iterable, Iterator
from warnings import warn

from IPython import get_ipython
from IPython.core.interactiveshell import InteractiveShell

__all__ = (
    "ui_events",
    "with_ui_events",
    "run_ui_poll_loop",
)


def _replay_events(shell: InteractiveShell, events: Iterable[Iterable]) -> None:
    kernel = shell.kernel  # type: ignore
    sys.stdout.flush()
    sys.stderr.flush()
    for stream, ident, parent in events:
        kernel.set_parent(ident, parent)
        # kernel._aborting does not exist on ipykernel 4
        if hasattr(kernel, "_aborting") and kernel._aborting:
            kernel._send_abort_reply(stream, parent, ident)
        else:
            kernel.execute_request(stream, ident, parent)


@contextmanager
def ui_events() -> Iterator[Callable]:
    """
    Gives you a function you can call to process ui events while running a long
    task inside a Jupyter cell.

    .. code-block: python
       with ui_events() as ui_poll:
          while some_condition:
             ui_poll(10)  # Process upto 10 UI events if any happened
             do_some_more_compute()


    - Delay processing `execute_request` IPython kernel events
    - Calls `kernel.do_one_iteration()`
    - Schedule replay of any blocked `execute_request` events upon
      exiting from the context manager
    """

    shell = get_ipython()
    kernel = shell.kernel
    events = []
    kernel.shell_handlers["execute_request"] = lambda *e: events.append(e)
    current_parent = (kernel._parent_ident, kernel._parent_header)

    def poll(n: int = 1) -> None:
        for _ in range(n):
            # ensure stdout still happens in the same cell
            kernel.set_parent(*current_parent)
            kernel.do_one_iteration()
            kernel.set_parent(*current_parent)

    try:
        poll()  # ensure poll is called at least once to correct output redirect
        yield poll
    finally:
        kernel.shell_handlers["execute_request"] = kernel.execute_request
        loop = asyncio.get_event_loop()
        if loop.is_running():
            loop.call_soon(lambda: _replay_events(shell, events))
        else:
            warn(
                "Automatic execution of scheduled cells only works with asyncio based ipython"
            )


def with_ui_events(its: Iterator, n: int = 1) -> Iterator:
    """
    Deal with kernel ui events while processing a long sequence

    :param its: Iterator to pass through
    :param n:   Number of events to process in between items

    - Delay processing `execute_request` IPython kernel events
    - Inject calls to `kernel.do_one_iteration()` in between iterations
    - Schedule replay of any blocked `execute_request` events when data sequence is exhausted
    """
    with ui_events() as poll:
        try:
            for x in its:
                poll(n)
                yield x
        except GeneratorExit:
            pass
        except Exception as e:
            raise e


def run_ui_poll_loop(f: Callable, sleep: float = 0.02, n: int = 1) -> Any:
    """
    Repeatedly call `f()` until it returns non-None value while also responding to widget events.

    This blocks execution of cells below in the notebook while still preserving
    interactivity of jupyter widgets.

    :param f: Function to periodically call (`f()` should not block for long)
    :param sleep: Amount of time to sleep in between polling (in seconds, 1/50 is the default)
    :param n: Number of events to process per iteration

    Returns
    =======
    First non-None value returned from `f()`
    """

    def as_iterator(f: Callable, sleep: float) -> Iterator:
        x = None
        while x is None:
            if sleep is not None:
                time.sleep(sleep)

            x = f()
            yield x

    for x in with_ui_events(as_iterator(f, sleep), n):
        if x is not None:
            return x



from html import escape
from pathlib import Path
from typing import Dict, Optional

from unfolded.map_sdk.environment import CURRENT_ENVIRONMENT, Environment
from unfolded.map_sdk.map.base import (
    BasemapParams,
    BaseNonInteractiveMap,
    MapInitialState,
    MapStyle,
    URLParams,
)
from unfolded.map_sdk.transport.html import HTMLTransport

TEMPLATE_DIR = (Path(__file__).parent / ".." / "templates").resolve()

IFRAME_TEMPLATE = """\
<iframe
    height="{height}"
    width="{width}"
    srcdoc="{srcdoc}"
></iframe>
"""


class HTMLMap(BaseNonInteractiveMap):
    transport: HTMLTransport
    style: MapStyle
    iframe: bool

    def __init__(
        self,
        initial_state: Optional[Dict] = None,
        style: Optional[Dict] = None,
        basemaps: Optional[Dict] = None,
        urls: Optional[Dict] = None,
        iframe: Optional[bool] = None,
    ):
        if CURRENT_ENVIRONMENT == Environment.DATABRICKS:
            self.iframe = True
        else:
            if iframe is not None:
                self.iframe = iframe
            else:
                self.iframe = False

        if initial_state:
            self.initial_state = MapInitialState(**initial_state).dict(
                by_alias=True, exclude_none=True
            )
        else:
            self.initial_state = {}

        if style:
            self.style = MapStyle(**style)
        else:
            self.style = MapStyle()

        if basemaps:
            validated_basemaps = BasemapParams(**basemaps)
            self.basemaps = validated_basemaps.dict(by_alias=True, exclude_none=True)
        else:
            self.basemaps = {}

        if urls:
            validated_urls = URLParams(**urls)
            self.urls = validated_urls.dict(by_alias=True, exclude_none=True)
        else:
            self.urls = {}

        self.transport = HTMLTransport()
        self.rendered = False

    def _repr_html_(self) -> str:
        return self.render()

    def render(self) -> str:
        html_string = self.transport.render_template(
            initial_state=self.initial_state,
            style=self.style.dict(exclude_none=True),
            basemaps=self.basemaps,
            urls=self.urls,
        )
        if not self.iframe:
            return html_string

        return IFRAME_TEMPLATE.format(
            height=self.style.height, width=self.style.width, srcdoc=escape(html_string)
        )



import sys
from typing import Any, Callable, Dict, List, Optional, Type, Union

from pydantic import Field, StrictBool, StrictStr

from unfolded.map_sdk.api.base import (
    Action,
    CamelCaseBaseModel,
    Number,
    Range,
    TimeRange,
)
from unfolded.map_sdk.api.enums import ActionType, FilterType
from unfolded.map_sdk.transport.base import (
    BaseInteractiveTransport,
    BaseNonInteractiveTransport,
    BaseTransport,
)
from unfolded.map_sdk.utils.validators import validate_kwargs

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class _PartialFilterSource(CamelCaseBaseModel):
    """Partially defined filter source where the datasetId can be left out."""

    data_id: Optional[StrictStr]
    """Identifier of the dataset that the filter applies to."""

    field_name: StrictStr
    """Field name to filter by.
    The field name can be retrieved as part of the dataset record.
    """


class FilterSource(_PartialFilterSource):
    """Source that the filter is applied to."""

    data_id: StrictStr
    """Identifier of the dataset that the filter applies to."""

    field_name: StrictStr
    """Field name to filter by. The field name can be retrieved as part of the dataset record.
    """


class _PartialBaseFilter(CamelCaseBaseModel):
    """Partial version of Filter for Filter creation"""

    id: Optional[StrictStr]
    """Unique identifier of the filter."""

    type: FilterType
    """Type of the filter."""

    # Incompatible types in assignment (expression has type "List[_T]", variable has type
    # "List[_PartialFilterSource]")  [assignment]
    sources: List[_PartialFilterSource] = Field(default_factory=list)
    """ Data source(s) to apply the filter to.
    note: Only TimeRangeFilter currently supports multiple sources.
    The first given source will be used for other filter types.
    """

    value: Any
    """Value to filter based on."""


class BaseFilter(_PartialBaseFilter):
    """Type encapsulating common filter properties."""

    id: StrictStr
    # Incompatible types in assignment (expression has type "List[FilterSource]", base class
    # "_PartialBaseFilter" defined the type as "List[_PartialFilterSource]")
    sources: List[FilterSource]  # type: ignore[assignment]


class _PartialRangeFilter(_PartialBaseFilter):
    """Partial RangeFilter for Filter creation"""

    type = FilterType.RANGE
    value: Range


class RangeFilter(BaseFilter):
    """Filter type that filters a range of values."""

    type = FilterType.RANGE
    value: Range


class _PartialSelectFilter(_PartialBaseFilter):
    """Partial SelectFilter for Filter creation"""

    type = FilterType.SELECT
    value: StrictBool


class SelectFilter(BaseFilter):
    """Filter type that filters a range of values."""

    type = FilterType.SELECT
    value: StrictBool


FilterView = Literal["side", "enlarged", "minified"]


class FilterTimelineUpdateProps(CamelCaseBaseModel):
    """A set of properties that can be updated on a timeline."""

    view: Optional[FilterView]
    """Current timeline presentation."""

    time_format: Optional[StrictStr]
    """Time format that the timeline is using in day.js supported format.
    https://day.js.org/docs/en/display/format
    """

    timezone: Optional[StrictStr]
    """Timezone that the timeline is using in tz format.
    https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    """

    is_animating: Optional[StrictBool]
    """Flag indicating whether the timeline is animating or not."""

    animation_speed: Optional[Number]
    """Speed at which timeline is animating."""


class FilterTimeline(FilterTimelineUpdateProps):
    """Time range filter properties that encapsulate timeline interaction."""

    view: FilterView
    """Current timeline presentation."""

    time_format: StrictStr
    """Time format that the timeline is using in day.js supported format.
    https://day.js.org/docs/en/display/format
    """

    timezone: Optional[StrictStr]
    """Timezone that the timeline is using in tz format.
    https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    """

    is_animating: StrictBool
    """Flag indicating whether the timeline is animating or not."""

    animation_speed: Number
    """Speed at which timeline is animating."""

    step: Number
    """Minimum animation step size in milliseconds."""


class _PartialTimeRangeFilter(_PartialBaseFilter):
    """Partial TimeRangeFilter for Filter creation"""

    type = FilterType.TIME_RANGE
    value: TimeRange
    domain: TimeRange
    timeline: FilterTimeline


class TimeRangeFilter(BaseFilter):

    type = FilterType.TIME_RANGE
    value: TimeRange
    domain: TimeRange
    timeline: FilterTimeline


class _PartialMultiSelectFilter(_PartialBaseFilter):
    """Partial MultiSelectFilter for Filter creation"""

    type = FilterType.MULTI_SELECT
    value: List[StrictStr]


class MultiSelectFilter(BaseFilter):
    """Filter type that filters a range of values."""

    type = FilterType.MULTI_SELECT
    value: List[StrictStr]


# Note that using a union like this as a pydantic type means that pydantic will try to match the
# input against each model in turn. This means that with this approach, we can't have any
# overlapping models definitinos, because there would be no way to know which filter is actually
# desired. Or we need to put the narrowest field first in the list.
#
# If that doesn't work, we would need to add a discriminator field like type
_PartialFilter = Union[
    _PartialRangeFilter,
    _PartialSelectFilter,
    _PartialTimeRangeFilter,
    _PartialMultiSelectFilter,
]
FilterCreationProps = _PartialFilter
FilterUpdateProps = _PartialFilter

PARTIAL_FILTER_CLASSES: Dict[FilterType, Type[_PartialFilter]] = {
    FilterType.RANGE: _PartialRangeFilter,
    FilterType.SELECT: _PartialSelectFilter,
    FilterType.TIME_RANGE: _PartialTimeRangeFilter,
    FilterType.MULTI_SELECT: _PartialMultiSelectFilter,
}


Filter = Union[RangeFilter, SelectFilter, TimeRangeFilter, MultiSelectFilter]


class _FilterTimelineUpdateProps(CamelCaseBaseModel):

    view: Optional[FilterView]
    """Current timeline presentation."""

    time_format: Optional[StrictStr]
    """Time format that the timeline is using in day.js supported format.
    https://day.js.org/docs/en/display/format
    """

    timezone: Optional[StrictStr]
    """Timezone that the timeline is using in tz format.
    https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    """

    is_animating: Optional[StrictBool]
    """Flag indicating whether the timeline is animating or not."""

    animation_speed: Optional[Number]
    """Speed at which timeline is animating."""


class FilterEventHandlers(CamelCaseBaseModel):
    on_filter_update: Optional[Callable[[Filter], None]]


###########
# ACTIONS #
###########


class GetFiltersAction(Action):
    """Action payload sent with `get_filters` calls"""

    type = ActionType.GET_FILTERS


class GetFilterByIdAction(Action):
    """Action payload sent with `get_filter_by_id` calls"""

    class Meta(Action.Meta):
        args = ["filter_id"]

    type = ActionType.GET_FILTER_BY_ID
    filter_id: StrictStr


class AddFilterAction(Action):
    class Meta(Action.Meta):
        args = ["filter"]

    type = ActionType.ADD_FILTER
    filter: FilterCreationProps


class UpdateFilterAction(Action):
    class Meta(Action.Meta):
        args = ["filter_id", "values"]

    type = ActionType.UPDATE_FILTER
    filter_id: StrictStr
    values: FilterUpdateProps


class RemoveFilterAction(Action):
    class Meta(Action.Meta):
        args = ["filter_id"]

    type = ActionType.REMOVE_FILTER
    filter_id: StrictStr


class UpdateTimelineAction(Action):
    class Meta(Action.Meta):
        args = ["filter_id", "values"]

    type = ActionType.UPDATE_TIMELINE
    filter_id: StrictStr
    values: _FilterTimelineUpdateProps


###########
# METHODS #
###########


class BaseFilterApiMethods:

    transport: BaseTransport

    @validate_kwargs(positional_only=["filter"])
    def add_filter(
        self,
        # pylint:disable = redefined-builtin
        filter: Union[FilterCreationProps, dict, None] = None,
        **kwargs: Any,
    ) -> Optional[Filter]:
        """Adds a new filter to the map.

        Args:
            filter (FilterCreationProps): The filter to add.

        Returns (widget map only):
            Filter: The filter that was added.
        """
        action = AddFilterAction(filter=filter if filter is not None else kwargs)

        # Fails mypy because Filter is a Union
        return self.transport.send_action_non_null(action=action, response_class=Filter)  # type: ignore[arg-type]

    @validate_kwargs(positional_only=["filter_id", "values"])
    def update_filter(
        self,
        filter_id: str,
        values: Union[FilterUpdateProps, dict, None] = None,
        **kwargs: Any,
    ) -> Optional[Filter]:
        """Updates an existing filter with given values.

        Args:
            filter_id (str): The id of the filter to update.
            values (Union[FilterUpdateProps, dict, None], optional): The new filter values. Defaults to None.

        Returns (widget map only):
            Filter: The updated filter.
        """
        action = UpdateFilterAction(
            filter_id=filter_id, values=values if values is not None else kwargs
        )

        # Fails mypy because Filter is a Union
        return self.transport.send_action_non_null(action=action, response_class=Filter)  # type: ignore[arg-type]

    @validate_kwargs(positional_only=["filter_id"])
    def remove_filter(self, filter_id: str) -> None:
        """Removes a filter from the map.

        Args:
            filter_id (str): The id of the filter to remove.

        Returns:
            None
        """
        action = RemoveFilterAction(filter_id=filter_id)
        self.transport.send_action(action=action, response_class=None)

    @validate_kwargs(positional_only=["filter_id", "values"])
    def update_timeline(
        self,
        filter_id: str,
        values: Union[FilterTimelineUpdateProps, dict, None] = None,
        **kwargs: Any,
    ) -> Optional[TimeRangeFilter]:
        """Updates a time range filter timeline with given values.

        Args:
            filter_id (str): The id of the time range filter to update.
            values (Union[FilterTimelineUpdateProps, dict, None], optional): The new timeline values. Defaults to None.

        Returns (widget map only):
            TimeRangeFilter: The updated time range filter.
        """
        action = UpdateTimelineAction(
            filter_id=filter_id, values=values if values is not None else kwargs
        )

        return self.transport.send_action_non_null(
            action=action, response_class=TimeRangeFilter
        )


class BaseInteractiveFilterApiMethods:

    transport: BaseInteractiveTransport

    def get_filters(self) -> List[Filter]:
        """Gets all the filters currently available in the map.

        Returns:
            List[Filter]: An array of filters.
        """
        action = GetFiltersAction()
        return self.transport.send_action_non_null(
            action=action, response_class=List[Filter]
        )

    @validate_kwargs(positional_only=["filter_id"])
    def get_filter_by_id(self, filter_id: str) -> Optional[Filter]:
        """Retrieves a filter by its identifier if it exists.

        Args:
            filter_id (str): Identifier of the filter to get.

        Returns:
            Optional[Filter]: Filter with a given identifier, or None if one doesn't exist.
        """
        action = GetFilterByIdAction(filter_id=filter_id)

        # Fails mypy because Filter is a Union
        return self.transport.send_action(action=action, response_class=Filter)  # type: ignore[arg-type]


class FilterApiNonInteractiveMixin(BaseFilterApiMethods):
    """Filter methods that are supported in non-interactive (i.e. pure HTML) maps"""

    transport: BaseNonInteractiveTransport

    def add_filter(
        self,
        # pylint:disable = redefined-builtin
        filter: Union[FilterCreationProps, dict, None] = None,
        **kwargs: Any,
    ) -> None:
        super().add_filter(filter, **kwargs)
        return

    @validate_kwargs(positional_only=["filter_id", "values"])
    def update_filter(
        self,
        filter_id: str,
        values: Union[FilterUpdateProps, dict, None] = None,
        **kwargs: Any,
    ) -> None:
        super().update_filter(filter_id, values, **kwargs)
        return

    @validate_kwargs(positional_only=["filter_id", "values"])
    def update_timeline(
        self,
        filter_id: str,
        values: Union[FilterTimelineUpdateProps, dict, None] = None,
        **kwargs: Any,
    ) -> None:
        super().update_timeline(filter_id, values, **kwargs)
        return


class FilterApiInteractiveMixin(BaseFilterApiMethods, BaseInteractiveFilterApiMethods):

    transport: BaseInteractiveTransport

    def add_filter(
        self,
        # pylint:disable = redefined-builtin
        filter: Union[FilterCreationProps, dict, None] = None,
        **kwargs: Any,
    ) -> Filter:
        return super().add_filter(filter, **kwargs)

    @validate_kwargs(positional_only=["filter_id", "values"])
    def update_filter(
        self,
        filter_id: str,
        values: Union[FilterUpdateProps, dict, None] = None,
        **kwargs: Any,
    ) -> Filter:
        return super().update_filter(filter_id, values, **kwargs)

    @validate_kwargs(positional_only=["filter_id", "values"])
    def update_timeline(
        self,
        filter_id: str,
        values: Union[FilterTimelineUpdateProps, dict, None] = None,
        **kwargs: Any,
    ) -> TimeRangeFilter:
        return super().update_timeline(filter_id, values, **kwargs)



from tests._utils import check_expected_comm_message
from tests.conftest import MockComm
from unfolded.map_sdk.map.widget import SyncWidgetMap


class TestGetDatasets:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_datasets()

        expected = {
            "type": "v1/map-sdk-get-datasets",
            "args": [],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestGetDatasetById:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_dataset_by_id("dataset-id")

        expected = {
            "type": "v1/map-sdk-get-dataset-by-id",
            "args": ["dataset-id"],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestAddDataset:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-add-dataset",
            "args": [
                {
                    "id": "dataset-id",
                    "type": "local",
                    "label": "dataset-label",
                    "data": "dataset data",
                }
            ],
            "options": {
                "autoCreateLayers": False,
                "centerMap": False,
            },
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        dataset = {
            "id": "dataset-id",
            "type": "local",
            "label": "dataset-label",
            "data": "dataset data",
        }
        auto_create_layers = False
        center_map = False

        mock_widget_map.add_dataset(
            dataset, auto_create_layers=auto_create_layers, center_map=center_map
        )

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        dataset = {
            "id": "dataset-id",
            "type": "local",
            "label": "dataset-label",
            "data": "dataset data",
        }
        auto_create_layers = False
        center_map = False

        mock_widget_map.add_dataset(
            **dataset, auto_create_layers=auto_create_layers, center_map=center_map
        )

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_raster_custom_stac_item(
        self,
        mock_comm: MockComm,
        sentinel2_stac_item: dict,
        mock_widget_map: SyncWidgetMap,
    ):
        raster_tile_dataset = {"type": "raster-tile", "metadata": sentinel2_stac_item}
        mock_widget_map.add_dataset(raster_tile_dataset)

        expected = {
            "type": "v1/map-sdk-add-dataset",
            "args": [{"type": "raster-tile", "metadata": sentinel2_stac_item}],
            "options": {"autoCreateLayers": True, "centerMap": True},
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)

    def test_raster_custom_stac_collection(
        self,
        mock_comm: MockComm,
        sentinel2_stac_collection: dict,
        mock_widget_map: SyncWidgetMap,
    ):
        raster_tile_dataset = {
            "type": "raster-tile",
            "metadata": sentinel2_stac_collection,
        }
        mock_widget_map.add_dataset(raster_tile_dataset)

        expected = {
            "type": "v1/map-sdk-add-dataset",
            "args": [{"type": "raster-tile", "metadata": sentinel2_stac_collection}],
            "options": {"autoCreateLayers": True, "centerMap": True},
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestAddTileDataset:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-add-tile-dataset",
            "args": [
                {
                    "id": "dataset-id",
                    "type": "vector-tile",
                    "label": "dataset-label",
                    "metadata": {
                        "metadataUrl": "http://example.com",
                    },
                }
            ],
            "options": {
                "autoCreateLayers": False,
                "centerMap": False,
            },
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        dataset = {
            "id": "dataset-id",
            "type": "vector-tile",
            "label": "dataset-label",
            "metadata": {
                "metadata_url": "http://example.com",
            },
        }
        auto_create_layers = False
        center_map = False

        mock_widget_map.add_tile_dataset(
            dataset, auto_create_layers=auto_create_layers, center_map=center_map
        )

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        dataset = {
            "id": "dataset-id",
            "type": "vector-tile",
            "label": "dataset-label",
            "metadata": {
                "metadata_url": "http://example.com",
            },
        }
        auto_create_layers = False
        center_map = False

        mock_widget_map.add_tile_dataset(
            **dataset, auto_create_layers=auto_create_layers, center_map=center_map
        )

        assert check_expected_comm_message(self.expected, mock_comm.log_send)


class TestUpdateDataset:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-update-dataset",
            "args": [
                "dataset-id",
                {
                    "label": "new-label",
                    "color": [0, 1, 0],
                },
            ],
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.update_dataset(
            "dataset-id",
            {
                "label": "new-label",
                "color": [0, 1, 0],
            },
        )

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        mock_widget_map.update_dataset(
            "dataset-id",
            **{
                "label": "new-label",
                "color": [0, 1, 0],
            },
        )

        assert check_expected_comm_message(self.expected, mock_comm.log_send)


class TestRemoveDataset:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.remove_dataset("dataset-id")

        expected = {
            "type": "v1/map-sdk-remove-dataset",
            "args": [
                "dataset-id",
            ],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)


class TestReplaceDataset:
    @property
    def expected(self):
        return {
            "type": "v1/map-sdk-replace-dataset",
            "args": [
                "dataset-id",
                {
                    "id": "dataset-id",
                    "type": "local",
                    "label": "dataset-label",
                    "data": "dataset data",
                },
            ],
            "options": {"force": False, "strict": False},
        }

    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.replace_dataset(
            "dataset-id",
            {
                "id": "dataset-id",
                "type": "local",
                "label": "dataset-label",
                "data": "dataset data",
            },
        )

        assert check_expected_comm_message(self.expected, mock_comm.log_send)

    def test_widget_message_kwargs(
        self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap
    ):
        mock_widget_map.replace_dataset(
            "dataset-id",
            **{
                "id": "dataset-id",
                "type": "local",
                "label": "dataset-label",
                "data": "dataset data",
            },
        )

        assert check_expected_comm_message(self.expected, mock_comm.log_send)


class TestGetDatasetWithData:
    def test_widget_message(self, mock_comm: MockComm, mock_widget_map: SyncWidgetMap):
        mock_widget_map.get_dataset_with_data("dataset-id")

        expected = {
            "type": "v1/map-sdk-get-dataset-with-data",
            "args": [
                "dataset-id",
            ],
        }

        assert check_expected_comm_message(expected, mock_comm.log_send)



from abc import abstractmethod
from typing import Any, List, Sequence, Union

from unfolded.map_sdk.api.base import Action
from unfolded.map_sdk.api.enums import EVENT_HANDLER_MAP, ActionType, EventType
from unfolded.map_sdk.api.filter_api import FilterEventHandlers
from unfolded.map_sdk.api.layer_api import LayerEventHandlers
from unfolded.map_sdk.api.map_api import MapEventHandlers
from unfolded.map_sdk.transport.base import BaseInteractiveTransport
from unfolded.map_sdk.utils.validators import validate_kwargs


class EventHandlers(MapEventHandlers, FilterEventHandlers, LayerEventHandlers):
    ...


###########
# ACTIONS #
###########


class SetEventHandlers(Action):
    class Meta(Action.Meta):
        args = ["names"]

    type: ActionType = ActionType.SET_EVENT_HANDLERS
    names: List[EventType]


class RemoveEventHandlers(Action):
    class Meta(Action.Meta):
        args = ["names"]

    type: ActionType = ActionType.REMOVE_EVENT_HANDLERS
    names: List[EventType]


###########
# METHODS #
###########


class BaseEventApiMethods:
    @abstractmethod
    def set_event_handlers(
        self, event_handlers: Union[dict, EventHandlers, None], **kwargs: Any
    ) -> None:
        ...

    @abstractmethod
    def remove_event_handlers(
        self, event_handlers: Sequence[Union[str, EventType]]
    ) -> None:
        ...


# The Event API is not implemented for the pure HTML static map because the static HTML bundle can't
# send events back to Python
class EventApiInteractiveMixin(BaseEventApiMethods):
    transport: BaseInteractiveTransport

    @validate_kwargs(positional_only=["event_handlers"])
    def set_event_handlers(
        self, event_handlers: Union[dict, EventHandlers, None] = None, **kwargs: Any
    ) -> None:
        if not isinstance(event_handlers, EventHandlers):
            event_handlers = (
                EventHandlers(**event_handlers)
                if event_handlers is not None
                else EventHandlers(**kwargs)
            )

        attribute_names = event_handlers.dict(exclude_none=True).keys()
        event_values: List[EventType] = []
        for attr_name in attribute_names:
            event_val = EVENT_HANDLER_MAP.get(attr_name)
            if event_val:
                event_values.append(event_val)

        action = SetEventHandlers(names=event_values)
        self.transport.send_action(action=action)

        self.transport.set_event_handlers(event_handlers.dict(exclude_none=True))

    def remove_event_handlers(
        self, event_handlers: Sequence[Union[str, EventType]]
    ) -> None:
        action = RemoveEventHandlers(names=event_handlers)
        self.transport.send_action(action=action)

        self.transport.remove_event_handlers(event_handlers)



from copy import deepcopy
from typing import List
from uuid import UUID

from tests.conftest import CommMessage


def check_expected_comm_message(expected: dict, actual: List[CommMessage]):

    new_items = deepcopy(actual)
    for new_item in new_items:
        data = new_item["data"]
        if data["method"] == "custom":
            content = data["content"]
            if "messageId" in content:
                UUID(content["messageId"])
                content.pop("messageId")  # type: ignore

    return any(item.get("data", {}).get("content", {}) == expected for item in new_items)  # type: ignore
